from pycram.orm.casts import StringType

from ormatic.custom_types import TypeType
from sqlalchemy import Boolean, Column, DateTime, Enum, Float, ForeignKey, Integer, MetaData, String, Table
from sqlalchemy.orm import RelationshipProperty, registry, relationship
import pycram.datastructures.dataclasses
import pycram.datastructures.grasp
import pycram.datastructures.pose
import pycram.designator
import pycram.designators.action_designator
import pycram.language
import pycram.orm.model
import pycram.plan

metadata = MetaData()


t_GraspDescription = Table(
    'GraspDescription', metadata,
    Column('id', Integer, primary_key=True),
    Column('approach_direction', Enum(pycram.datastructures.enums.Grasp), nullable=False),
    Column('vertical_alignment', Enum(pycram.datastructures.enums.Grasp)),
    Column('rotate_gripper', Boolean, nullable=False)
)

t_Header = Table(
    'Header', metadata,
    Column('id', Integer, primary_key=True),
    Column('frame_id', String, nullable=False),
    Column('stamp', DateTime, nullable=False),
    Column('sequence', Integer, nullable=False)
)

t_ORMActionNode = Table(
    'ORMActionNode', metadata,
    Column('id', Integer, primary_key=True)
)

t_ORMMotionNode = Table(
    'ORMMotionNode', metadata,
    Column('id', Integer, primary_key=True)
)

t_PlanNode = Table(
    'PlanNode', metadata,
    Column('id', Integer, primary_key=True),
    Column('status', Enum(pycram.datastructures.enums.TaskStatus), nullable=False),
    Column('start_time', DateTime),
    Column('end_time', DateTime),
    Column('polymorphic_type', String)
)

t_Quaternion = Table(
    'Quaternion', metadata,
    Column('id', Integer, primary_key=True),
    Column('x', Float, nullable=False),
    Column('y', Float, nullable=False),
    Column('z', Float, nullable=False),
    Column('w', Float, nullable=False)
)

t_Vector3 = Table(
    'Vector3', metadata,
    Column('id', Integer, primary_key=True),
    Column('x', Float, nullable=False),
    Column('y', Float, nullable=False),
    Column('z', Float, nullable=False)
)

t_Pose = Table(
    'Pose', metadata,
    Column('id', Integer, primary_key=True),
    Column('position_id', ForeignKey('Vector3.id'), nullable=False),
    Column('orientation_id', ForeignKey('Quaternion.id'), nullable=False),
    Column('polymorphic_type', String)
)

t_PoseStamped = Table(
    'PoseStamped', metadata,
    Column('id', Integer, primary_key=True),
    Column('pose_id', ForeignKey('Pose.id'), nullable=False),
    Column('header_id', ForeignKey('Header.id'), nullable=False),
    Column('polymorphic_type', String)
)

t_Transform = Table(
    'Transform', metadata,
    Column('id', ForeignKey('Pose.id'), primary_key=True)
)

t_ActionDescription = Table(
    'ActionDescription', metadata,
    Column('id', Integer, primary_key=True),
    Column('robot_position_id', ForeignKey('PoseStamped.id')),
    Column('robot_torso_height', Float),
    Column('polymorphic_type', String)
)

t_FrozenObject = Table(
    'FrozenObject', metadata,
    Column('id', Integer, primary_key=True),
    Column('name', String, nullable=False),
    Column('concept', StringType, nullable=False),
    Column('pose_id', ForeignKey('PoseStamped.id'))
)

t_GraspPose = Table(
    'GraspPose', metadata,
    Column('id', ForeignKey('PoseStamped.id'), primary_key=True),
    Column('arm', Enum(pycram.datastructures.enums.Arms), nullable=False),
    Column('grasp_description_id', ForeignKey('GraspDescription.id'), nullable=False)
)

t_TransformStamped = Table(
    'TransformStamped', metadata,
    Column('id', ForeignKey('PoseStamped.id'), primary_key=True),
    Column('pose_id', ForeignKey('Transform.id'), nullable=False),
    Column('child_frame_id', String, nullable=False)
)

t_CloseAction = Table(
    'CloseAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('arm', Enum(pycram.datastructures.enums.Arms), nullable=False),
    Column('grasping_prepose_distance', Float, nullable=False)
)

t_DetectAction = Table(
    'DetectAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('technique', Enum(pycram.datastructures.enums.DetectionTechnique), nullable=False),
    Column('state', Enum(pycram.datastructures.enums.DetectionState)),
    Column('region', String),
    Column('object_at_execution_id', ForeignKey('FrozenObject.id'))
)

t_FaceAtAction = Table(
    'FaceAtAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('pose_id', ForeignKey('PoseStamped.id'), nullable=False),
    Column('keep_joint_states', Boolean, nullable=False)
)

t_GraspingAction = Table(
    'GraspingAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('arm', Enum(pycram.datastructures.enums.Arms), nullable=False),
    Column('object_at_execution_id', ForeignKey('FrozenObject.id'), nullable=False)
)

t_GripAction = Table(
    'GripAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('gripper', Enum(pycram.datastructures.enums.Arms), nullable=False),
    Column('effort', Float, nullable=False),
    Column('object_at_execution_id', ForeignKey('FrozenObject.id'))
)

t_LanguageNode = Table(
    'LanguageNode', metadata,
    Column('id', ForeignKey('PlanNode.id'), primary_key=True),
    Column('action_id', ForeignKey('ActionDescription.id'), nullable=False)
)

t_LookAtAction = Table(
    'LookAtAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('target_id', ForeignKey('PoseStamped.id'), nullable=False)
)

t_MoveTorsoAction = Table(
    'MoveTorsoAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('torso_state', Enum(pycram.datastructures.enums.TorsoState), nullable=False)
)

t_NavigateAction = Table(
    'NavigateAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('target_location_id', ForeignKey('PoseStamped.id'), nullable=False),
    Column('keep_joint_states', Boolean, nullable=False)
)

t_ORMResolvedActionNode = Table(
    'ORMResolvedActionNode', metadata,
    Column('id', Integer, primary_key=True),
    Column('designator_ref_id', ForeignKey('ActionDescription.id'), nullable=False)
)

t_OpenAction = Table(
    'OpenAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('arm', Enum(pycram.datastructures.enums.Arms), nullable=False),
    Column('grasping_prepose_distance', Float, nullable=False)
)

t_ParkArmsAction = Table(
    'ParkArmsAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('arm', Enum(pycram.datastructures.enums.Arms), nullable=False)
)

t_PickUpAction = Table(
    'PickUpAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('arm', Enum(pycram.datastructures.enums.Arms), nullable=False),
    Column('object_at_execution_id', ForeignKey('FrozenObject.id'))
)

t_PlaceAction = Table(
    'PlaceAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('arm', Enum(pycram.datastructures.enums.Arms), nullable=False),
    Column('target_location_id', ForeignKey('PoseStamped.id'), nullable=False),
    Column('object_at_execution_id', ForeignKey('FrozenObject.id'))
)

t_ReachToPickUpAction = Table(
    'ReachToPickUpAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('arm', Enum(pycram.datastructures.enums.Arms), nullable=False),
    Column('grasp_description_id', ForeignKey('GraspDescription.id'), nullable=False),
    Column('object_at_execution_id', ForeignKey('FrozenObject.id'))
)

t_ReleaseAction = Table(
    'ReleaseAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('gripper', Enum(pycram.datastructures.enums.Arms), nullable=False),
    Column('object_at_execution_id', ForeignKey('FrozenObject.id'))
)

t_SearchAction = Table(
    'SearchAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('target_location_id', ForeignKey('PoseStamped.id'), nullable=False),
    Column('object_type', TypeType)
)

t_SetGripperAction = Table(
    'SetGripperAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('gripper', Enum(pycram.datastructures.enums.Arms), nullable=False),
    Column('motion', Enum(pycram.datastructures.enums.GripperState), nullable=False)
)

t_TaskTreeNode = Table(
    'TaskTreeNode', metadata,
    Column('id', Integer, primary_key=True),
    Column('start_time', DateTime, nullable=False),
    Column('end_time', DateTime),
    Column('status', Enum(pycram.datastructures.enums.TaskStatus), nullable=False),
    Column('reason', String),
    Column('action_id', ForeignKey('ActionDescription.id')),
    Column('parent_id', ForeignKey('TaskTreeNode.id'))
)

t_TransportAction = Table(
    'TransportAction', metadata,
    Column('id', ForeignKey('ActionDescription.id'), primary_key=True),
    Column('arm', Enum(pycram.datastructures.enums.Arms), nullable=False),
    Column('target_location_id', ForeignKey('PoseStamped.id'), nullable=False),
    Column('object_at_execution_id', ForeignKey('FrozenObject.id'))
)

t_CodeNode = Table(
    'CodeNode', metadata,
    Column('id', ForeignKey('LanguageNode.id'), primary_key=True)
)

t_ParallelNode = Table(
    'ParallelNode', metadata,
    Column('id', ForeignKey('LanguageNode.id'), primary_key=True)
)

t_SequentialNode = Table(
    'SequentialNode', metadata,
    Column('id', ForeignKey('LanguageNode.id'), primary_key=True)
)

t_MonitorNode = Table(
    'MonitorNode', metadata,
    Column('id', ForeignKey('SequentialNode.id'), primary_key=True)
)

t_RepeatNode = Table(
    'RepeatNode', metadata,
    Column('id', ForeignKey('SequentialNode.id'), primary_key=True),
    Column('repeat', Integer, nullable=False)
)

t_TryAllNode = Table(
    'TryAllNode', metadata,
    Column('id', ForeignKey('ParallelNode.id'), primary_key=True)
)

t_TryInOrderNode = Table(
    'TryInOrderNode', metadata,
    Column('id', ForeignKey('SequentialNode.id'), primary_key=True)
)

mapper_registry = registry(metadata=metadata)

m_GraspDescription = mapper_registry.map_imperatively(pycram.datastructures.grasp.GraspDescription, t_GraspDescription, )

m_Vector3 = mapper_registry.map_imperatively(pycram.datastructures.pose.Vector3, t_Vector3, )

m_Quaternion = mapper_registry.map_imperatively(pycram.datastructures.pose.Quaternion, t_Quaternion, )

m_Pose = mapper_registry.map_imperatively(pycram.datastructures.pose.Pose, t_Pose, properties = dict(position=relationship('Vector3',foreign_keys=[t_Pose.c.position_id]), 
orientation=relationship('Quaternion',foreign_keys=[t_Pose.c.orientation_id])), polymorphic_on = "polymorphic_type", polymorphic_identity = "Pose")

m_Header = mapper_registry.map_imperatively(pycram.datastructures.pose.Header, t_Header, )

m_PoseStamped = mapper_registry.map_imperatively(pycram.datastructures.pose.PoseStamped, t_PoseStamped, properties = dict(pose=relationship('Pose',foreign_keys=[t_PoseStamped.c.pose_id]), 
header=relationship('Header',foreign_keys=[t_PoseStamped.c.header_id])), polymorphic_on = "polymorphic_type", polymorphic_identity = "PoseStamped")

m_ActionDescription = mapper_registry.map_imperatively(pycram.designator.ActionDescription, t_ActionDescription, properties = dict(robot_position=relationship('PoseStamped',foreign_keys=[t_ActionDescription.c.robot_position_id])), polymorphic_on = "polymorphic_type", polymorphic_identity = "ActionDescription")

m_PlanNode = mapper_registry.map_imperatively(pycram.plan.PlanNode, t_PlanNode, polymorphic_on = "polymorphic_type", polymorphic_identity = "PlanNode")

m_TaskTreeNode = mapper_registry.map_imperatively(pycram.orm.model.TaskTreeNode, t_TaskTreeNode, properties = dict(action=relationship('ActionDescription',foreign_keys=[t_TaskTreeNode.c.action_id]), 
parent=relationship('TaskTreeNode',foreign_keys=[t_TaskTreeNode.c.parent_id])))

m_FrozenObject = mapper_registry.map_imperatively(pycram.datastructures.dataclasses.FrozenObject, t_FrozenObject, properties = dict(pose=relationship('PoseStamped',foreign_keys=[t_FrozenObject.c.pose_id]), 
concept=t_FrozenObject.c.concept))

m_ORMActionNode = mapper_registry.map_imperatively(pycram.plan.ActionNode, t_ORMActionNode, )

m_ORMMotionNode = mapper_registry.map_imperatively(pycram.plan.MotionNode, t_ORMMotionNode, )

m_ORMResolvedActionNode = mapper_registry.map_imperatively(pycram.plan.ResolvedActionNode, t_ORMResolvedActionNode, properties = dict(designator_ref=relationship('ActionDescription',foreign_keys=[t_ORMResolvedActionNode.c.designator_ref_id])))

m_Transform = mapper_registry.map_imperatively(pycram.datastructures.pose.Transform, t_Transform, polymorphic_identity = "Transform", inherits = m_Pose)

m_TransformStamped = mapper_registry.map_imperatively(pycram.datastructures.pose.TransformStamped, t_TransformStamped, properties = dict(pose=relationship('Transform',foreign_keys=[t_TransformStamped.c.pose_id])), polymorphic_identity = "TransformStamped", inherits = m_PoseStamped)

m_GraspPose = mapper_registry.map_imperatively(pycram.datastructures.pose.GraspPose, t_GraspPose, properties = dict(grasp_description=relationship('GraspDescription',foreign_keys=[t_GraspPose.c.grasp_description_id])), polymorphic_identity = "GraspPose", inherits = m_PoseStamped)

m_MoveTorsoAction = mapper_registry.map_imperatively(pycram.designators.action_designator.MoveTorsoAction, t_MoveTorsoAction, polymorphic_identity = "MoveTorsoAction", inherits = m_ActionDescription)

m_SetGripperAction = mapper_registry.map_imperatively(pycram.designators.action_designator.SetGripperAction, t_SetGripperAction, polymorphic_identity = "SetGripperAction", inherits = m_ActionDescription)

m_ParkArmsAction = mapper_registry.map_imperatively(pycram.designators.action_designator.ParkArmsAction, t_ParkArmsAction, polymorphic_identity = "ParkArmsAction", inherits = m_ActionDescription)

m_NavigateAction = mapper_registry.map_imperatively(pycram.designators.action_designator.NavigateAction, t_NavigateAction, properties = dict(target_location=relationship('PoseStamped',foreign_keys=[t_NavigateAction.c.target_location_id])), polymorphic_identity = "NavigateAction", inherits = m_ActionDescription)

m_LookAtAction = mapper_registry.map_imperatively(pycram.designators.action_designator.LookAtAction, t_LookAtAction, properties = dict(target=relationship('PoseStamped',foreign_keys=[t_LookAtAction.c.target_id])), polymorphic_identity = "LookAtAction", inherits = m_ActionDescription)

m_OpenAction = mapper_registry.map_imperatively(pycram.designators.action_designator.OpenAction, t_OpenAction, polymorphic_identity = "OpenAction", inherits = m_ActionDescription)

m_CloseAction = mapper_registry.map_imperatively(pycram.designators.action_designator.CloseAction, t_CloseAction, polymorphic_identity = "CloseAction", inherits = m_ActionDescription)

m_FaceAtAction = mapper_registry.map_imperatively(pycram.designators.action_designator.FaceAtAction, t_FaceAtAction, properties = dict(pose=relationship('PoseStamped',foreign_keys=[t_FaceAtAction.c.pose_id])), polymorphic_identity = "FaceAtAction", inherits = m_ActionDescription)

m_SearchAction = mapper_registry.map_imperatively(pycram.designators.action_designator.SearchAction, t_SearchAction, properties = dict(target_location=relationship('PoseStamped',foreign_keys=[t_SearchAction.c.target_location_id]), 
object_type=t_SearchAction.c.object_type), polymorphic_identity = "SearchAction", inherits = m_ActionDescription)

m_DetectAction = mapper_registry.map_imperatively(pycram.designators.action_designator.DetectAction, t_DetectAction, properties = dict(object_at_execution=relationship('FrozenObject',foreign_keys=[t_DetectAction.c.object_at_execution_id])), polymorphic_identity = "DetectAction", inherits = m_ActionDescription)

m_GraspingAction = mapper_registry.map_imperatively(pycram.designators.action_designator.GraspingAction, t_GraspingAction, properties = dict(object_at_execution=relationship('FrozenObject',foreign_keys=[t_GraspingAction.c.object_at_execution_id])), polymorphic_identity = "GraspingAction", inherits = m_ActionDescription)

m_ReleaseAction = mapper_registry.map_imperatively(pycram.designators.action_designator.ReleaseAction, t_ReleaseAction, properties = dict(object_at_execution=relationship('FrozenObject',foreign_keys=[t_ReleaseAction.c.object_at_execution_id])), polymorphic_identity = "ReleaseAction", inherits = m_ActionDescription)

m_GripAction = mapper_registry.map_imperatively(pycram.designators.action_designator.GripAction, t_GripAction, properties = dict(object_at_execution=relationship('FrozenObject',foreign_keys=[t_GripAction.c.object_at_execution_id])), polymorphic_identity = "GripAction", inherits = m_ActionDescription)

m_ReachToPickUpAction = mapper_registry.map_imperatively(pycram.designators.action_designator.ReachToPickUpAction, t_ReachToPickUpAction, properties = dict(grasp_description=relationship('GraspDescription',foreign_keys=[t_ReachToPickUpAction.c.grasp_description_id]), 
object_at_execution=relationship('FrozenObject',foreign_keys=[t_ReachToPickUpAction.c.object_at_execution_id])), polymorphic_identity = "ReachToPickUpAction", inherits = m_ActionDescription)

m_PickUpAction = mapper_registry.map_imperatively(pycram.designators.action_designator.PickUpAction, t_PickUpAction, properties = dict(object_at_execution=relationship('FrozenObject',foreign_keys=[t_PickUpAction.c.object_at_execution_id])), polymorphic_identity = "PickUpAction", inherits = m_ActionDescription)

m_PlaceAction = mapper_registry.map_imperatively(pycram.designators.action_designator.PlaceAction, t_PlaceAction, properties = dict(target_location=relationship('PoseStamped',foreign_keys=[t_PlaceAction.c.target_location_id]), 
object_at_execution=relationship('FrozenObject',foreign_keys=[t_PlaceAction.c.object_at_execution_id])), polymorphic_identity = "PlaceAction", inherits = m_ActionDescription)

m_TransportAction = mapper_registry.map_imperatively(pycram.designators.action_designator.TransportAction, t_TransportAction, properties = dict(target_location=relationship('PoseStamped',foreign_keys=[t_TransportAction.c.target_location_id]), 
object_at_execution=relationship('FrozenObject',foreign_keys=[t_TransportAction.c.object_at_execution_id])), polymorphic_identity = "TransportAction", inherits = m_ActionDescription)

m_LanguageNode = mapper_registry.map_imperatively(pycram.language.LanguageNode, t_LanguageNode, properties = dict(action=relationship('ActionDescription',foreign_keys=[t_LanguageNode.c.action_id])), polymorphic_identity = "LanguageNode", inherits = m_PlanNode)

m_SequentialNode = mapper_registry.map_imperatively(pycram.language.SequentialNode, t_SequentialNode, polymorphic_identity = "SequentialNode", inherits = m_LanguageNode)

m_ParallelNode = mapper_registry.map_imperatively(pycram.language.ParallelNode, t_ParallelNode, polymorphic_identity = "ParallelNode", inherits = m_LanguageNode)

m_CodeNode = mapper_registry.map_imperatively(pycram.language.CodeNode, t_CodeNode, polymorphic_identity = "CodeNode", inherits = m_LanguageNode)

m_TryInOrderNode = mapper_registry.map_imperatively(pycram.language.TryInOrderNode, t_TryInOrderNode, polymorphic_identity = "TryInOrderNode", inherits = m_SequentialNode)

m_MonitorNode = mapper_registry.map_imperatively(pycram.language.MonitorNode, t_MonitorNode, polymorphic_identity = "MonitorNode", inherits = m_SequentialNode)

m_RepeatNode = mapper_registry.map_imperatively(pycram.language.RepeatNode, t_RepeatNode, polymorphic_identity = "RepeatNode", inherits = m_SequentialNode)

m_TryAllNode = mapper_registry.map_imperatively(pycram.language.TryAllNode, t_TryAllNode, polymorphic_identity = "TryAllNode", inherits = m_ParallelNode)



import rosgraph
import rosnode
import rospy
# import rospkg

from rospkg import RosPack, ResourceNotFound
from typing_extensions import Any


def get_node_names(namespace=None):
    return rosnode.get_node_names(namespace)


def create_ros_pack(ros_paths: Any = None) -> RosPack:
    """
    Creates a RosPack instance to search for resources of ros packages.

    :param ros_paths: An ordered list of paths to search for resources.
    :return: An instance of RosPack
    """
    return RosPack(ros_paths)


def get_ros_package_path(package_name: str) -> str:
    rospack = create_ros_pack()
    return rospack.get_path(package_name)


def get_parameter(name: str) -> Any:
    return rospy.get_param(name)


def wait_for_message(topic_name: str, msg_type: Any):
    return rospy.wait_for_message(topic_name, msg_type)


def is_master_online():
    return rosgraph.is_master_online()


def sleep(duration: float):
    rospy.sleep(duration)


def get_time():
    return rospy.get_time()


def create_timer(duration: rospy.Duration, callback, oneshot=False):
    return rospy.Timer(duration, callback, oneshot=oneshot)



from __future__ import annotations

import rospy
from rospy.logger_level_service_caller import LoggerLevelServiceCaller
import inspect
from pathlib import Path

from typing_extensions import TYPE_CHECKING

if TYPE_CHECKING:
    from ..datastructures.enums import LoggerLevel


PYCRAM_LOGGER_NAME = "pycram"
logger_level_service_caller = LoggerLevelServiceCaller()



def _get_caller_method_name():
    """
    Get the name of the method that called the function from which this function is called. It is intended as a helper
    function for the log functions.

    :return: Name of the method that called the function from which this function is called.
    """
    return inspect.stack()[2][3]


def _get_caller_method_line():
    """
    Get the line of the method that called the function from which this function is called. It is intended as a helper
    function for the log functions.

    :return: Line number of the method that called the function from which this function is called.
    """
    return inspect.stack()[2][2]


def _get_caller_file_name():
    """
    Get the file name of the method that called the function from which this function is called. It is intended as a helper
    function for the log functions.

    :return: File name of the method that called the function from which this function is called.
    """
    path = Path(inspect.stack()[2][1])
    return path.name


def set_logger_level(level: LoggerLevel):
    """
    Set the logger level for the pycram logger.

    :param level: The level to set the logger to, possible values are: rospy.DEBUG, rospy.INFO, rospy.WARN, rospy.ERROR,
    rospy.FATAL.
    """
    logger_level_service_caller.get_loggers(rospy.get_name())
    logger_level_service_caller.send_logger_change_message(rospy.get_name(),
                                                           f"rosout.{PYCRAM_LOGGER_NAME}",
                                                           level.value)


def logwarn(message: str):
    rospy.logwarn(f"[{_get_caller_file_name()}:{_get_caller_method_line()}:{_get_caller_method_name()}] {message}",
                  logger_name=PYCRAM_LOGGER_NAME)


def loginfo(message: str):
    rospy.loginfo(f"[{_get_caller_file_name()}:{_get_caller_method_line()}:{_get_caller_method_name()}] {message}",
                  logger_name=PYCRAM_LOGGER_NAME)


def logerr(message: str):
    rospy.logerr(f"[{_get_caller_file_name()}:{_get_caller_method_line()}:{_get_caller_method_name()}] {message}",
                 logger_name=PYCRAM_LOGGER_NAME)


def logdebug(message: str):
    rospy.logdebug(f"[{_get_caller_file_name()}:{_get_caller_method_line()}:{_get_caller_method_name()}] {message}",
                   logger_name=PYCRAM_LOGGER_NAME)


def logwarn_once(message: str):
    rospy.logwarn_once(f"[{_get_caller_file_name()}:{_get_caller_method_line()}:{_get_caller_method_name()}] {message}",
                       logger_name=PYCRAM_LOGGER_NAME)


def loginfo_once(message: str):
    rospy.loginfo_once(f"[{_get_caller_file_name()}:{_get_caller_method_line()}:{_get_caller_method_name()}] {message}",
                       logger_name=PYCRAM_LOGGER_NAME)


def logerr_once(message: str):
    rospy.logerr_once(f"[{_get_caller_file_name()}:{_get_caller_method_line()}:{_get_caller_method_name()}] {message}",
                      logger_name=PYCRAM_LOGGER_NAME)


def logdebug_once(message: str):
    rospy.logdebug_once(
        f"[{_get_caller_file_name()}:{_get_caller_method_line()}:{_get_caller_method_name()}] {message}",
        logger_name=PYCRAM_LOGGER_NAME)



from threading import Lock

import numpy as np
from .default_process_modules import *

from ..datastructures.world import World
from ..datastructures.enums import ExecutionType, ObjectType
from ..designator import ObjectDesignatorDescription
from ..designators.motion_designator import MoveMotion, DetectingMotion, MoveTCPMotion, MoveArmJointsMotion, \
    MoveJointsMotion, MoveGripperMotion, OpeningMotion, ClosingMotion
from ..local_transformer import LocalTransformer
from ..process_module import ProcessModuleManager, ProcessModule
from .default_process_modules import DefaultOpen, DefaultClose, DefaultMoveGripper, DefaultMoveJoints, DefaultMoveTCP, \
    DefaultNavigation, DefaultMoveHead, DefaultDetecting, DefaultMoveArmJoints, DefaultWorldStateDetecting, \
    DefaultDetectingReal
from ..robot_description import RobotDescription
from ..ros import logdebug
from ..external_interfaces import giskard
from ..world_concepts.world_object import Object
from ..external_interfaces.robokudo import send_query


class TiagoNavigationReal(ProcessModule):
    def _execute(self, designator: MoveMotion):
        logdebug(f"Sending goal to giskard to Move the robot")
        giskard.achieve_cartesian_goal(designator.target, RobotDescription.current_robot_description.base_link, "map")


class TiagoMoveHeadReal(ProcessModule):
    def _execute(self, designator: MoveMotion):
        target = designator.target
        robot = World.robot

        local_transformer = LocalTransformer()

        pan_link = RobotDescription.current_robot_description.kinematic_chains["neck"].links[0]
        tilt_link = RobotDescription.current_robot_description.kinematic_chains["neck"].links[1]

        pan_joint = RobotDescription.current_robot_description.kinematic_chains["neck"].joints[0]
        tilt_joint = RobotDescription.current_robot_description.kinematic_chains["neck"].joints[1]
        pose_in_pan = local_transformer.transform_pose(target, robot.get_link_tf_frame(pan_link))
        pose_in_tilt = local_transformer.transform_pose(target, robot.get_link_tf_frame(tilt_link))

        new_pan = np.arctan2(pose_in_pan.position.y, pose_in_pan.position.x)
        new_tilt = np.arctan2(pose_in_tilt.position.z, pose_in_tilt.position.x ** 2 + pose_in_tilt.position.y ** 2) * -1

        current_pan = robot.get_joint_position(pan_joint)
        current_tilt = robot.get_joint_position(tilt_joint)

        giskard.avoid_all_collisions()
        giskard.achieve_joint_goal({pan_link: new_pan + current_pan,
                                    tilt_link: new_tilt + current_tilt})


class TiagoDetectingReal(ProcessModule):
    def _execute(self, designator: DetectingMotion):
        query_result = send_query(ObjectDesignatorDescription(types=[designator.object_type]))
        # print(query_result)
        obj_pose = query_result["ClusterPoseBBAnnotator"]

        lt = LocalTransformer()
        obj_pose = lt.transform_pose(obj_pose, World.robot.get_link_tf_frame("torso_lift_link"))
        obj_pose.orientation = [0, 0, 0, 1]
        obj_pose.position.x += 0.05

        world_obj = World.current_world.get_object_by_type(designator.object_type)
        if world_obj:
            world_obj[0].set_pose(obj_pose)
            return world_obj[0]
        elif designator.object_type == ObjectType.JEROEN_CUP:
            cup = Object("cup", ObjectType.JEROEN_CUP, "jeroen_cup.stl", pose=obj_pose)
            return cup
        elif designator.object_type == ObjectType.BOWL:
            bowl = Object("bowl", ObjectType.BOWL, "bowl.stl", pose=obj_pose)
            return bowl

        return world_obj[0]


class TiagoMoveTCPReal(ProcessModule):
    def _execute(self, designator: MoveTCPMotion):
        lt = LocalTransformer()
        pose_in_map = lt.transform_pose(designator.target, "map")

        if designator.allow_gripper_collision:
            giskard.allow_gripper_collision(designator.arm)
        giskard.achieve_cartesian_goal(pose_in_map, RobotDescription.current_robot_description.get_arm_chain(
            designator.arm).get_tool_frame(),
                                       "torso_lift_link")


class TiagoManager(DefaultManager):

    def __init__(self):
        super().__init__()
        self.robot_name = "tiago_dual"

    def navigate(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultNavigation(self._navigate_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return TiagoNavigationReal(self._navigate_lock)

    def looking(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveHead(self._looking_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return TiagoMoveHeadReal(self._looking_lock)

    def move_tcp(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveTCP(self._move_tcp_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return TiagoMoveTCPReal(self._move_tcp_lock)

    def world_state_detecting(self):
        if (ProcessModuleManager.execution_type == ExecutionType.SIMULATED or
                ProcessModuleManager.execution_type == ExecutionType.REAL):
            return DefaultWorldStateDetecting(self._world_state_detecting_lock)

    def move_gripper(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveGripper(self._move_gripper_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultMoveGripperReal(self._move_gripper_lock)

TiagoManager()


import os
import time
import unittest
from datetime import timedelta

from .datastructures.world import UseProspectionWorld
from .worlds.bullet_world import BulletWorld
from .world_concepts.world_object import Object
from .datastructures.pose import PoseStamped
from .robot_description import RobotDescription, RobotDescriptionManager
from .process_module import ProcessModule
from .datastructures.enums import WorldMode
from .object_descriptors.urdf import ObjectDescription
from .plan import Plan
from .ros import loginfo
from pycrap.ontologies import Milk, Robot, Kitchen, Cereal
try:
    from .ros_utils.viz_marker_publisher import VizMarkerPublisher
except ImportError:
    loginfo("Could not import VizMarkerPublisher. This is probably because you are not running ROS.")

class EmptyBulletWorldTestCase(unittest.TestCase):
    """
    Base class for unit tests that require and ordinary setup and teardown of the empty bullet-world.
    """

    world: BulletWorld
    #viz_marker_publisher: VizMarkerPublisher
    extension: str = ObjectDescription.get_file_extension()
    render_mode = WorldMode.DIRECT

    @classmethod
    def setUpClass(cls):
        cls.world = BulletWorld(mode=cls.render_mode)
        if "ROS_VERSION" in os.environ:
            cls.viz_marker_publisher = VizMarkerPublisher()

    def setUp(self):
        self.world.reset_world(remove_saved_states=True)
        Plan.current_plan = None
        with UseProspectionWorld():
            pass


    def tearDown(self):
        time.sleep(0.05)
        self.world.reset_world(remove_saved_states=True)
        with UseProspectionWorld():
            pass

    @classmethod
    def tearDownClass(cls):
        if "ROS_VERSION" in os.environ:
            cls.viz_marker_publisher._stop_publishing()
        cls.world.exit()


class BulletWorldTestCase(EmptyBulletWorldTestCase):
    """
    Class for unit tests that require a bullet-world with a PR2, kitchen, milk and cereal.
    """

    world: BulletWorld
    #viz_marker_publisher: VizMarkerPublisher
    extension: str = ObjectDescription.get_file_extension()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        rdm = RobotDescriptionManager()
        rdm.load_description("pr2")
        cls.milk = Object("milk", Milk, "milk.stl", pose=PoseStamped.from_list([1.3, 1, 0.9]))
        cls.robot = Object(RobotDescription.current_robot_description.name, Robot,
                           RobotDescription.current_robot_description.name + cls.extension)
        cls.kitchen = Object("kitchen", Kitchen, "kitchen" + cls.extension)
        cls.cereal = Object("cereal", Cereal, "breakfast_cereal.stl",
                            pose=PoseStamped.from_list([1.3, 0.7, 0.95]))

class BulletWorldGUITestCase(BulletWorldTestCase):
    render_mode = WorldMode.GUI



import os
from pathlib import Path

from pycram.testing import BulletWorldTestCase
from pycram.object_descriptors.urdf import ObjectDescription as URDFObject
from pycram.config import world_conf as conf


class TestCacheManager(BulletWorldTestCase):

    def test_generate_description_and_write_to_cache(self):
        cache_manager = self.world.cache_manager
        path = os.path.join(self.world.conf.resources_path, "objects/apartment.urdf")
        extension = Path(path).suffix
        cache_path = os.path.join(self.world.conf.cache_dir, "apartment.urdf")
        apartment = URDFObject(path)
        apartment.generate_description_from_file(path, "apartment", extension, cache_path)
        self.assertTrue(cache_manager.is_cached(path, apartment))



from typing import Optional, Tuple

from trimesh import Trimesh
from typing_extensions import List, Any, Union, Dict, Self, TYPE_CHECKING

from .urdf import ObjectDescription as UrdfObjectDescription
from ..config.world_conf import WorldConfig
from ..datastructures.dataclasses import VisualShape, BoxVisualShape, Color, AxisAlignedBoundingBox, RotatedBoundingBox, \
    BoundingBox
from ..datastructures.enums import JointType
from ..datastructures.pose import PoseStamped, Point
from ..description import JointDescription as AbstractJointDescription, LinkDescription as AbstractLinkDescription, \
    ObjectDescription as AbstractObjectDescription, ObjectDescription

if TYPE_CHECKING:
    pass


class NamedBoxVisualShape(BoxVisualShape):
    def __init__(self, name: str, color: Color, visual_frame_position: List[float], half_extents: List[float]):
        super().__init__(color, visual_frame_position, half_extents)
        self._name: str = name

    @property
    def name(self) -> str:
        return self._name


class LinkDescription(AbstractLinkDescription):

    def __init__(self, name: str, visual_frame_position: List[float], half_extents: List[float],
                 color: Color = Color()):
        super().__init__(NamedBoxVisualShape(name, color, visual_frame_position, half_extents))

    @property
    def geometry(self) -> Union[VisualShape, None]:
        return self.parsed_description

    @property
    def origin(self) -> PoseStamped:
        return PoseStamped.from_list(self.parsed_description.visual_frame_position)

    @property
    def name(self) -> str:
        return self.parsed_description.name

    @property
    def color(self) -> Color:
        return self.parsed_description.rgba_color


class JointDescription(AbstractJointDescription):

    @property
    def parent(self) -> str:
        raise NotImplementedError

    @property
    def child(self) -> str:
        raise NotImplementedError

    @property
    def type(self) -> JointType:
        return JointType.UNKNOWN

    @property
    def axis(self) -> Point:
        return Point(x=0, y=0, z=0)

    @property
    def has_limits(self) -> bool:
        return False

    @property
    def lower_limit(self) -> Union[float, None]:
        return 0

    @property
    def upper_limit(self) -> Union[float, None]:
        return 0

    @property
    def parent_link_name(self) -> str:
        raise NotImplementedError

    @property
    def child_link_name(self) -> str:
        raise NotImplementedError

    @property
    def origin(self) -> PoseStamped:
        raise NotImplementedError

    @property
    def name(self) -> str:
        raise NotImplementedError


class ObjectDescription(AbstractObjectDescription):
    """
    A generic description of an object in the environment. This description can be applied to any object.
    The current use case involves perceiving objects using RoboKudo and spawning them with specified size and color.
    """

    class Link(AbstractObjectDescription.Link, LinkDescription):
        ...

    class RootLink(AbstractObjectDescription.RootLink, Link):
        ...

    class Joint(AbstractObjectDescription.Joint, JointDescription):
        ...

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._links = [LinkDescription(*args, **kwargs)]

    def merge_description(self, other: ObjectDescription, parent_link: Optional[str] = None,
                          child_link: Optional[str] = None,
                          joint_type: JointType = JointType.FIXED,
                          axis: Optional[Point] = None,
                          lower_limit: Optional[float] = None, upper_limit: Optional[float] = None,
                          child_pose_wrt_parent: Optional[PoseStamped] = None,
                          in_place: bool = False,
                          new_description_file: Optional[str] = None) -> Union[ObjectDescription, Self]:
        return self.merge_using_bounding_boxes(other, child_pose_wrt_parent, new_description_file)

    def merge_using_bounding_boxes(self, other: ObjectDescription, child_pose_wrt_parent: Optional[PoseStamped] = None,
                                   new_description_file: Optional[str] = None) -> Union[ObjectDescription, Self]:
        """
        Merge the current object description with another object description by merging their bounding boxes and
        creating a URDF description with the new merged bounding box/ mesh.

        :param other: The other object description to merge with.
        :param child_pose_wrt_parent: The pose of the child object wrt the parent object.
        :param new_description_file: The path to save the new description file.
        :return: The new object description.
        """
        my_bounding_box: AxisAlignedBoundingBox = self._links[0].geometry.get_axis_aligned_bounding_box()
        other_bounding_box: AxisAlignedBoundingBox = other._links[0].geometry.get_axis_aligned_bounding_box()
        transform = child_pose_wrt_parent.to_transform_stamped(self.name)
        other_bounding_box: RotatedBoundingBox = other_bounding_box.get_rotated_box(transform)
        new_mesh = BoundingBox.merge_multiple_bounding_boxes_into_mesh([my_bounding_box, other_bounding_box],
                                                                       use_random_events=False)
        return self.create_urdf_from_mesh(new_mesh, path=new_description_file)

    def merge_using_urdf(self, other: ObjectDescription, child_pose_wrt_parent: Optional[PoseStamped] = None,
                         new_description_file: Optional[str] = None) -> Union[ObjectDescription, Self]:
        """
        Merge the current object description with another object description by creating a URDF description for both
        objects and merging them.

        :param other: The other object description to merge with.
        :param child_pose_wrt_parent: The pose of the child object wrt the parent object.
        :param new_description_file: The path to save the new description file.
        :return: The new object description.
        """
        my_urdf_description = self.create_urdf_from_mesh()
        other_urdf_description = other.create_urdf_from_mesh()
        return my_urdf_description.merge_description(other_urdf_description,
                                                     child_pose_wrt_parent=child_pose_wrt_parent,
                                                     new_description_file=new_description_file)

    def create_urdf_from_mesh(self, mesh: Optional[Trimesh] = None,
                              mesh_file_path: Optional[str] = None,
                              path: Optional[str] = None) -> UrdfObjectDescription:
        """
        Create a URDF description from the mesh and the current object description.

        :param mesh: The mesh to create the URDF description from.
        :param mesh_file_path: The path to the mesh file.
        :param path: The path to save the URDF description.
        :return: The URDF object description.
        """
        mesh = mesh if mesh else self.get_mesh()
        mesh_file_path = mesh_file_path if mesh_file_path else f"{WorldConfig.cache_dir}/{self.name}.obj"
        mesh.export(mesh_file_path)
        path = f"{WorldConfig.cache_dir}/{self.name}.urdf" if not path else path
        new_description = UrdfObjectDescription()
        new_description.generate_from_mesh_file(mesh_file_path, self.name, path)
        new_description.update_description_from_file(path)
        return new_description

    def get_mesh(self) -> Trimesh:
        """
        Get the mesh of the object.
        """
        return self._links[0].geometry.get_axis_aligned_bounding_box().as_mesh

    def load_description(self, path: str) -> Any:
        ...

    @classmethod
    def generate_from_mesh_file(cls, path: str, name: str, save_path: str, color: Color) -> str:
        raise NotImplementedError

    @classmethod
    def generate_from_description_file(cls, path: str, save_path: str, make_mesh_paths_absolute: bool = True) -> str:
        raise NotImplementedError

    @classmethod
    def generate_from_parameter_server(cls, name: str, save_path: str) -> str:
        raise NotImplementedError

    @property
    def parent_map(self) -> Dict[str, Tuple[str, str]]:
        return {}

    @property
    def link_map(self) -> Dict[str, LinkDescription]:
        return {self._links[0].name: self._links[0]}

    @property
    def joint_map(self) -> Dict[str, JointDescription]:
        return {}

    @property
    def child_map(self) -> Dict[str, List[Tuple[str, str]]]:
        return {}

    def add_joint(self, name: str, child: str, joint_type: JointType,
                  axis: Point, parent: Optional[str] = None, origin: Optional[PoseStamped] = None,
                  lower_limit: Optional[float] = None, upper_limit: Optional[float] = None,
                  is_virtual: Optional[bool] = False) -> None:
        ...

    @property
    def shape_data(self) -> List[float]:
        return self._links[0].geometry.shape_data()['halfExtents']

    @property
    def color(self) -> Color:
        return self._links[0].color

    @property
    def links(self) -> List[LinkDescription]:
        return self._links

    def get_link_by_name(self, link_name: str) -> LinkDescription:
        if link_name == self._links[0].name:
            return self._links[0]

    @property
    def joints(self) -> List[JointDescription]:
        return []

    def get_joint_by_name(self, joint_name: str) -> JointDescription:
        ...

    def get_root(self) -> str:
        return self._links[0].name

    def get_chain(self, start_link_name: str, end_link_name: str, joints: Optional[bool] = True,
                  links: Optional[bool] = True, fixed: Optional[bool] = True) -> List[str]:
        raise NotImplementedError("Do Not Do This on generic objects as they have no chains")

    @staticmethod
    def get_file_extension() -> str:
        raise NotImplementedError("Do Not Do This on generic objects as they have no extensions")

    @property
    def origin(self) -> PoseStamped:
        return self._links[0].origin

    @property
    def name(self) -> str:
        return self._links[0].name



from __future__ import annotations

import numpy as np
from typing_extensions import Union, List, Optional, TYPE_CHECKING, Self

from ..datastructures.enums import JointType
from ..datastructures.pose import TransformStamped, PoseStamped, Point

if TYPE_CHECKING:
    from ..description import Link


class AbstractConstraint:
    """
    Represents an abstract constraint concept, this could be used to create joints for example or any kind of constraint
    between two links in the world.
    """

    def __init__(self,
                 parent_link: Link,
                 child_link: Link,
                 _type: JointType,
                 parent_to_constraint: TransformStamped,
                 child_to_constraint: TransformStamped):
        self.parent_link: Link = parent_link
        self.child_link: Link = child_link
        self.type: JointType = _type
        self.parent_to_constraint = parent_to_constraint
        self.child_to_constraint = child_to_constraint
        self._parent_to_child = None

    def get_child_object_pose(self) -> PoseStamped:
        """
        :return: The pose of the child object.
        """
        return self.child_link.object.pose

    def get_child_object_pose_given_parent(self, pose: PoseStamped) -> PoseStamped:
        """
        Get the pose of the child object given the parent pose.

        :param pose: The parent object pose.
        :return: The pose of the child object.
        """
        pose = self.parent_link.get_pose_given_object_pose(pose)
        child_link_pose = self.get_child_link_target_pose_given_parent(pose)
        return self.child_link.get_object_pose_given_link_pose(child_link_pose)

    def set_child_link_pose(self):
        """
        Set the target pose of the child object to the current pose of the child object in the parent object frame.
        """
        self.child_link.set_object_pose_given_link_pose(self.get_child_link_target_pose())

    def get_child_link_target_pose(self) -> PoseStamped:
        """
        :return: The target pose of the child object. (The pose of the child object in the parent object frame)
        """
        return self.parent_to_child_transform.to_pose_stamped()

    def get_child_link_target_pose_given_parent(self, parent_pose: PoseStamped) -> PoseStamped:
        """
        Get the target pose of the child object link given the parent link pose.

        :param parent_pose: The parent link pose.
        :return: The target pose of the child object link.
        """
        return (parent_pose.to_transform_stamped(self.parent_link.tf_frame) * self.parent_to_child_transform).to_pose_stamped()

    @property
    def parent_to_child_transform(self) -> TransformStamped:
        """
        Return the transform from the parent link to the child link of the constraint.

        :return: The transform from the parent link to the child link of the constraint.
        """
        if self._parent_to_child is None:
            if self.parent_to_constraint is not None and self.child_to_constraint is not None:
                self._parent_to_child = ~self.parent_to_constraint * self.child_to_constraint
        return self._parent_to_child

    @parent_to_child_transform.setter
    def parent_to_child_transform(self, transform: TransformStamped) -> None:
        self._parent_to_child = transform

    @property
    def parent_object_id(self) -> int:
        """
        :return: The id of the parent object of the constraint
        """
        return self.parent_link.object_id

    @property
    def child_object_id(self) -> int:
        """
        :return: The id of the child object of the constraint
        """
        return self.child_link.object_id

    @property
    def parent_link_id(self) -> int:
        """
        :return: The id of the parent link of the constraint
        """
        return self.parent_link.id

    @property
    def child_link_id(self) -> int:
        """
        :return: The id of the child link of the constraint
        """
        return self.child_link.id

    @property
    def position_wrt_parent_as_list(self) -> List[float]:
        """
        :return: The constraint frame pose with respect to the parent origin as a list
        """
        return self.pose_wrt_parent.position.to_list()

    @property
    def orientation_wrt_parent_as_list(self) -> List[float]:
        """
        :return: The constraint frame orientation with respect to the parent origin as a list
        """
        return self.pose_wrt_parent.orientation.to_list()

    @property
    def pose_wrt_parent(self) -> PoseStamped:
        """
        :return: The joint frame pose with respect to the parent origin
        """
        return self.parent_to_constraint.to_pose_stamped()

    @property
    def position_wrt_child_as_list(self) -> List[float]:
        """
        :return: The constraint frame pose with respect to the child origin as a list
        """
        return self.pose_wrt_child.position.to_list()

    @property
    def orientation_wrt_child_as_list(self) -> List[float]:
        """
        :return: The constraint frame orientation with respect to the child origin as a list
        """
        return self.pose_wrt_child.orientation.to_list()

    @property
    def pose_wrt_child(self) -> PoseStamped:
        """
        :return: The joint frame pose with respect to the child origin
        """
        return self.child_to_constraint.to_pose_stamped()


class Constraint(AbstractConstraint):
    """
    Represents a constraint between two links in the World.
    """

    def __init__(self,
                 parent_link: Link,
                 child_link: Link,
                 _type: JointType,
                 axis_in_child_frame: Point,
                 constraint_to_parent: TransformStamped,
                 child_to_constraint: TransformStamped):
        parent_to_constraint = ~constraint_to_parent
        AbstractConstraint.__init__(self, parent_link, child_link, _type, parent_to_constraint, child_to_constraint)
        self.axis: Point = axis_in_child_frame

    @property
    def axis_as_list(self) -> List[float]:
        """
        :return: The axis of this constraint as a list of xyz
        """
        return [self.axis.x, self.axis.y, self.axis.z]


class Attachment(AbstractConstraint):
    def __init__(self,
                 parent_link: Link,
                 child_link: Link,
                 bidirectional: bool = False,
                 parent_to_child_transform: Optional[TransformStamped] = None,
                 constraint_id: Optional[int] = None,
                 is_inverse: bool = False):
        """
        Creates an attachment between the parent object link and the child object link.
        This could be a bidirectional attachment, meaning that both objects will move when one moves.

        :param parent_link: The parent object link.
        :param child_link: The child object link.
        :param bidirectional: If true, both objects will move when one moves.
        :param parent_to_child_transform: The transform from the parent link to the child object link.
        :param constraint_id: The id of the constraint in the simulator.
        """
        super().__init__(parent_link, child_link, JointType.FIXED, parent_to_child_transform,
                         TransformStamped.from_list(frame=child_link.tf_frame))
        self.id = constraint_id
        self.bidirectional: bool = bidirectional
        self._loose: bool = False
        self.is_inverse: bool = is_inverse

        if parent_to_child_transform is not None:
            self.parent_to_child_transform = parent_to_child_transform

        elif self.parent_to_child_transform is None:
            self.update_transform()

        if self.id is None:
            self.add_fixed_constraint()

    @property
    def parent_object(self):
        return self.parent_link.object

    @property
    def child_object(self):
        return self.child_link.object

    def update_transform_and_constraint(self) -> None:
        """
        Update the transform and constraint of this attachment.
        """
        self.update_transform()
        self.update_constraint()

    def update_transform(self) -> None:
        """
        Update the transform of this attachment by calculating the transform from the parent link to the child link.
        """
        self.parent_to_child_transform = self.calculate_transform()

    def update_constraint(self) -> None:
        """
        Update the constraint of this attachment by removing the old constraint if one exists and adding a new one.
        """
        self.remove_constraint_if_exists()
        self.add_fixed_constraint()

    def add_fixed_constraint(self) -> None:
        """
        Add a fixed constraint between the parent link and the child link.
        """
        self.id = self.parent_link.add_fixed_constraint_with_link(self.child_link,
                                                                  ~self.parent_to_child_transform)

    def calculate_transform(self) -> TransformStamped:
        """
        Calculate the transform from the parent link to the child link.
        """
        return self.parent_link.get_transform_to_link(self.child_link)

    def remove_constraint_if_exists(self) -> None:
        """
        Remove the constraint between the parent and the child links if one exists.
        """
        if self.child_link in self.parent_link.constraint_ids.keys():
            self.parent_link.remove_constraint_with_link(self.child_link)

    def get_inverse(self) -> 'Attachment':
        """
        :return: A new Attachment object with the parent and child links swapped.
        """
        attachment = Attachment(self.child_link, self.parent_link, self.bidirectional,
                                constraint_id=self.id, is_inverse=not self.is_inverse)
        attachment.loose = not self._loose
        return attachment

    @property
    def loose(self) -> bool:
        """
        If true, then the child object will not move when parent moves.
        """
        return self._loose

    @loose.setter
    def loose(self, loose: bool) -> None:
        """
        Set the loose property of this attachment.

        :param loose: If true, then the child object will not move when parent moves.
        """
        self._loose = loose and not self.bidirectional

    def __del__(self) -> None:
        """
        Remove the constraint between the parent and the child links if one exists when the attachment is deleted.
        """
        self.remove_constraint_if_exists()

    def __copy__(self):
        return Attachment(self.parent_link, self.child_link, self.bidirectional, self.parent_to_child_transform,
                          self.id)

    def __eq__(self, other):
        return (self.parent_link.name == other.parent_link.name
                and self.child_link.name == other.child_link.name
                and self.bidirectional == other.bidirectional
                and self.loose == other.loose
                and np.allclose(self.parent_to_child_transform.translation.to_list(),
                                other.parent_to_child_transform.translation.to_list(), rtol=0, atol=1e-3)
                and np.allclose(self.parent_to_child_transform.rotation.to_list(),
                                other.parent_to_child_transform.rotation.to_list(), rtol=0, atol=1e-3))

    def __hash__(self):
        return hash((self.parent_link.name, self.child_link.name, self.bidirectional, self.parent_to_child_transform))



# used for delayed evaluation of typing until python 3.11 becomes mainstream
from __future__ import annotations

from typing_extensions import Callable, List, Optional, Any


class Event:
    """
    Base implementation of events in PyCRAM.
    Events allow to attach handler methods to events that fire for specific occurences in the world.

    :ivar handler: List of methods that are called when this event is fired.
    """

    def __init__(self):
        self.handlers: List[Callable] = []

    def add(self, handler: Callable) -> None:
        """
        Adds a new handler to the list of handlers. All handler methods are called when this event is fired.
        Handler have to take the event sender as parameter as well as args* which can contain further parameter.

        :param handler: A method that should be added
        """
        self.handlers.append(handler)

    def remove(self, handler: Callable) -> None:
        """
        Removes a method from the list of handlers, the method will not be called when the event is fired.

        :param handler: The method that should be removed.
        """
        self.handlers.remove(handler)

    def fire(self, sender: Any, earg: Optional[Any] = None) -> None:
        """
        Fire this event, this causes every method to be called with a sender as well as additional args.

        :param sender: The entity that fired the event.
        :param earg: Additional arguments.
        """
        for handler in self.handlers:
            handler(sender, earg)

    def __iadd__(self, other: Callable) -> Event:
        """
        Operator overload that allows to add handlers by the '+=' operator.

        :param other: The handler that should be added.
        :return: This instance
        """
        self.add(other)
        return self

    def __isub__(self, other: Callable) -> Event:
        """
        Operator overload that allows to remove methods as handlers by using the '-=' operator.

        :param other: The method that should be removed as handler.
        :return: This instance
        """
        self.remove(other)
        return self

    __call__ = fire
    """
    Allows to directly call the reference.
    """


import numpy as np
from threading import Lock
from typing_extensions import Any

from .default_process_modules import *
from ..datastructures.enums import ExecutionType
from ..external_interfaces.tmc import tmc_gripper_control, tmc_talk
from ..robot_description import RobotDescription
from ..process_module import ProcessModule
from ..local_transformer import LocalTransformer
from ..designators.motion_designator import *
from ..external_interfaces import giskard
from ..datastructures.world import World
from ..ros import  logdebug
try:
    from pydub import AudioSegment
    from pydub.playback import play
    from gtts import gTTS
except ImportError:
    logdebug("pydub and gtts are not installed. HSR speech simulated will not work.")

import io




###########################################################
########## Process Modules for the Real HSRB ###############
###########################################################


class HSRBNavigationReal(ProcessModule):
    """
    Process module for the real HSRB that sends a cartesian goal to giskard to move the robot base
    """

    def _execute(self, designator: MoveMotion) -> Any:
        logdebug(f"Sending goal to giskard to Move the robot")
        # giskard.achieve_cartesian_goal(designator.target, robot_description.base_link, "map")
        # todome fix this
        # queryPoseNav(designator.target)


class HSRBMoveHeadReal(ProcessModule):
    """
    Process module for the real HSRB that sends a pose goal to giskard to move the robot head
    """

    def _execute(self, desig: LookingMotion):
        target = desig.target
        giskard.move_head_to_pose(target)


class HSRBMoveTCPReal(ProcessModule):
    """
    Moves the tool center point of the real HSRB while avoiding all collisions via giskard
    """

    def _execute(self, designator: MoveTCPMotion) -> Any:
        lt = LocalTransformer()
        pose_in_map = lt.transform_pose(designator.target, "map")
        giskard.avoid_all_collisions()
        if designator.allow_gripper_collision:
            giskard.allow_gripper_collision(designator.arm)
        giskard.achieve_cartesian_goal(pose_in_map, RobotDescription.current_robot_description.get_arm_chain(
            designator.arm).get_tool_frame(), "map")


class HSRBMoveArmJointsReal(ProcessModule):
    """
    Moves the arm joints of the real HSRB to the given configuration while avoiding all collisions via giskard
    """

    def _execute(self, designator: MoveArmJointsMotion) -> Any:
        joint_goals = {}
        if designator.left_arm_poses:
            joint_goals.update(designator.left_arm_poses)
        giskard.avoid_all_collisions()
        giskard.achieve_joint_goal(joint_goals)


class HSRBMoveJointsReal(ProcessModule):
    """
    Moves any joint using giskard, avoids all collisions while doint this.
    """

    def _execute(self, designator: MoveJointsMotion) -> Any:
        name_to_position = dict(zip(designator.names, designator.positions))
        giskard.avoid_all_collisions()
        giskard.achieve_joint_goal(name_to_position)


class HSRBMoveGripperReal(ProcessModule):
    """
     Opens or closes the gripper of the real HSRB with the help of giskard.
     """

    def _execute(self, designator: MoveGripperMotion) -> Any:
        tmc_gripper_control(designator)


class HSRBOpenReal(ProcessModule):
    """
    This process Modules tries to open an already grasped container via giskard
    """

    def _execute(self, designator: OpeningMotion) -> Any:
        giskard.achieve_open_container_goal(
            RobotDescription.current_robot_description.get_arm_chain(designator.arm).get_tool_frame(),
            designator.object_part.name)


class HSRBCloseReal(ProcessModule):
    """
    This process module executes close a an already grasped container via giskard
    """

    def _execute(self, designator: ClosingMotion) -> Any:
        giskard.achieve_close_container_goal(
            RobotDescription.current_robot_description.get_arm_chain(designator.arm).get_tool_frame(),
            designator.object_part.name)


class HSRBTalkReal(ProcessModule):
    """
    Let the robot speak over tmc interface.
    """

    def _execute(self, designator: TalkingMotion) -> Any:
        tmc_talk(designator)


###########################################################
########## Process Modules for the Semi Real HSRB ###############
###########################################################
class HSRBNavigationSemiReal(ProcessModule):
    """
    Process module for the real HSRB that sends a cartesian goal to giskard to move the robot base
    """

    def _execute(self, designator: MoveMotion) -> Any:
        logdebug(f"Sending goal to giskard to Move the robot")
        giskard.teleport_robot(designator.target)


class HSRBTalkSemiReal(ProcessModule):
    """
    Low Level implementation to let the robot talk using gTTS and pydub.
    """

    def _execute(self, designator: TalkingMotion) -> Any:
        """
        Convert text to speech using gTTS, modify the pitch and play it without saving to disk.
        """
        sentence = designator.cmd
        # Create a gTTS object
        tts = gTTS(text=sentence, lang='en', slow=False)

        # Save the speech to an in-memory file
        mp3_fp = io.BytesIO()
        tts.write_to_fp(mp3_fp)
        mp3_fp.seek(0)

        # Load the audio into pydub from the in-memory file
        audio = AudioSegment.from_file(mp3_fp, format="mp3")

        # Speed up the audio slightly
        faster_audio = audio.speedup(playback_speed=1.2)

        # Play the modified audio
        play(faster_audio)


###########################################################
########## HSRB MANAGER ###############
###########################################################
class HSRBManager(DefaultManager):

    def __init__(self):
        super().__init__()
        self.robot_name = "hsrb"
        self._navigate_lock = Lock()


    def navigate(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultNavigation(self._navigate_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return HSRBNavigationReal(self._navigate_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.SEMI_REAL:
            return HSRBNavigationSemiReal(self._navigate_lock)

    def looking(self):
        if ProcessModuleManager.execution_type == ExecutionType.REAL:
            return HSRBMoveHeadReal(self._looking_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.SEMI_REAL:
            return HSRBMoveHeadReal(self._looking_lock)

    def detecting(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultDetecting(self._detecting_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultDetectingReal(self._detecting_lock)

    def move_tcp(self):
        if  ProcessModuleManager.execution_type == ExecutionType.REAL:
            return HSRBMoveTCPReal(self._move_tcp_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.SEMI_REAL:
            return HSRBMoveTCPReal(self._move_tcp_lock)

    def move_arm_joints(self):
        if ProcessModuleManager.execution_type == ExecutionType.REAL:
            return HSRBMoveArmJointsReal(self._move_arm_joints_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.SEMI_REAL:
            return HSRBMoveArmJointsReal(self._move_arm_joints_lock)

    def move_joints(self):
        if ProcessModuleManager.execution_type == ExecutionType.REAL:
            return HSRBMoveJointsReal(self._move_joints_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.SEMI_REAL:
            return HSRBMoveJointsReal(self._move_joints_lock)

    def move_gripper(self):
        if ProcessModuleManager.execution_type == ExecutionType.REAL:
            return HSRBMoveGripperReal(self._move_gripper_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.SEMI_REAL:
            return HSRBMoveGripperReal(self._move_gripper_lock)

    def open(self):
        if ProcessModuleManager.execution_type == ExecutionType.REAL:
            return HSRBOpenReal(self._open_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.SEMI_REAL:
            return HSRBOpenReal(self._open_lock)

    def close(self):
        if ProcessModuleManager.execution_type == ExecutionType.REAL:
            return HSRBCloseReal(self._close_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.SEMI_REAL:
            return HSRBCloseReal(self._close_lock)

    def talk(self):
        if ProcessModuleManager.execution_type == ExecutionType.REAL:
            return HSRBTalkReal(self._talk_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.SEMI_REAL:
            return HSRBTalkSemiReal(self._talk_lock)

HSRBManager()



from abc import abstractmethod

from scipy.signal import butter, lfilter


class Filter:
    """
    Abstract class to ensure that every supported filter needs to implement the filter method.

    :method filter: Abstract method to filter the given data.
    """

    @abstractmethod
    def filter(self, data):
        pass


class Butterworth(Filter):
    """
    Implementation for a Butterworth filter.

    :param order: The order of the filter (default is 4).
    :param cutoff: The cutoff frequency of the filter (default is 10).
    :param fs: The sampling frequency of the data (default is 60).
  """

    def __init__(self, order=4, cutoff=10, fs=60):
        self.order = order
        self.cutoff = cutoff
        self.fs = fs

        self.b, self.a = butter(self.order, cutoff / (0.5 * fs), btype='low')

    def filter(self, data: list):
        """
        Filters the given data using a Butterworth filter.

        :param data: The data to be filtered.

        :return: The filtered data.
        """
        return lfilter(self.b, self.a, data)



# used for delayed evaluation of typing until python 3.11 becomes mainstream
from __future__ import annotations

import dataclasses
import glob
import importlib
from os.path import dirname, basename, isfile, join
import math
from dataclasses import field
from enum import Enum
from itertools import product

import numpy as np
from scipy.spatial.transform import Rotation as R
from typing_extensions import List, Dict, Union, Optional, Tuple, TYPE_CHECKING

from .datastructures.dataclasses import VirtualMobileBaseJoints, ManipulatorData, Rotations
from .datastructures.enums import Arms, Grasp, GripperState, GripperType, JointType, DescriptionType, StaticJointState
from .datastructures.pose import GraspDescription, PoseStamped
from .helper import parse_mjcf_actuators, find_multiverse_resources_path, \
    get_robot_description_path
from .object_descriptors.urdf import ObjectDescription as URDFObject
from .ros import logerr
from .tf_transformations import quaternion_multiply
from .utils import suppress_stdout_stderr

if TYPE_CHECKING:
    from .datastructures.pose import Pose

class RobotDescriptionManager:
    """
    Singleton class to manage multiple robot descriptions. Stores all robot descriptions and loads a robot description
    according to the name of the loaded robot.
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls, *args, **kwargs)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        """
        Initialize the RobotDescriptionManager, if no instance exists a new instance is created.
        """
        if self._initialized: return
        self.descriptions: Dict[str, RobotDescription] = {}
        self._initialized = True
        self.register_all_descriptions()

    def load_description(self, name: str):
        """
        Loads a robot description according to the name of the robot. This required that a robot description with the
        corresponding name is registered.

        :param name: Name of the robot to which the description should be loaded.
        :return: The loaded robot description.
        """
        if RobotDescription.current_robot_description:
            RobotDescription.current_robot_description.unload()

        for key in self.descriptions.keys():
            if name in self.descriptions[key].urdf_object.name or key in name.lower():
                self.descriptions[key].load()
                return self.descriptions[key]
        logerr(f"Robot description {name} not found")

    def register_description(self, description: RobotDescription):
        """
        Register a robot description to the RobotDescriptionManager. The description is stored with the name of the
        description as key. This will later be used to load the description.

        :param description: RobotDescription to register.
        """
        if description.name in self.descriptions.keys():
            raise ValueError(f"Description {description.name} already exists")
        self.descriptions[description.name] = description

    @staticmethod
    def register_all_descriptions():
        modules = glob.glob(join(dirname(__file__) + "/robot_descriptions", "*.py"))
        __all__ = [basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py')]

        for module_name in __all__:
            try:
                importlib.import_module(f".{module_name}", package="pycram.robot_descriptions")
            except Exception as e:
                print(f"Error loading module {module_name}: {e}")


class RobotDescription:
    """
    Base class of a robot description. Contains all necessary information about a robot, like the URDF, the base link,
    the torso link and joint, the kinematic chains and cameras.
    """
    current_robot_description: RobotDescription = None
    """
    The currently loaded robot description.
    """
    name: str
    """
    Name of the robot
    """
    base_link: str
    """
    Base link of the robot
    """
    torso_link: str
    """
    Torso link of the robot
    """
    torso_joint: str
    """
    Torso joint of the robot
    """
    urdf_object: URDFObject
    """
    Parsed URDF of the robot
    """
    kinematic_chains: Dict[str, KinematicChainDescription]
    """
    All kinematic chains defined for this robot
    """
    cameras: Dict[str, CameraDescription]
    """
    All cameras defined for this robot
    """
    links: List[str]
    """
    All links defined in the URDF
    """
    joints: List[str]
    """
    All joints defined in the URDF, by default fixed joints are not included
    """
    virtual_mobile_base_joints: Optional[VirtualMobileBaseJoints] = None
    """
    Virtual mobile base joint names for mobile robots, these joints are not part of the URDF, however they are used to
    move the robot in the simulation (e.g. set_pose for the robot would actually move these joints)
    """
    gripper_name: Optional[str] = None
    """
    Name of the gripper of the robot if it has one, this is used when the gripper is a different Object with its own
    description file outside the robot description file.
    """
    neck: Dict[str, List[str]]
    """
    Dictionary of neck links and joints. Keys are yaw, pitch and roll, values are [link, joint]
    """

    def __init__(self, name: str, base_link: str, torso_link: str, torso_joint: str, urdf_path: str,
                 virtual_mobile_base_joints: Optional[VirtualMobileBaseJoints] = None, mjcf_path: Optional[str] = None,
                 ignore_joints: Optional[List[str]] = None,
                 gripper_name: Optional[str] = None):
        """
        Initialize the RobotDescription. The URDF is loaded from the given path and used as basis for the kinematic
        chains.

        :param name: Name of the robot
        :param base_link: Base link of the robot, meaning the first link in the URDF
        :param torso_link: Torso link of the robot
        :param torso_joint: Torso joint of the robot, this is the joint that moves the torso upwards if there is one
        :param urdf_path: Path to the URDF file of the robot
        :param virtual_mobile_base_joints: Virtual mobile base joint names for mobile robots
        :param mjcf_path: Path to the MJCF file of the robot
        :param ignore_joints: List of joint names that are not used.
        :param gripper_name: Name of the gripper of the robot if it has one and is a separate Object.
        """
        self.name = name
        self.base_link = base_link
        self.torso_link = torso_link
        self.torso_joint = torso_joint
        self.ignore_joints = ignore_joints if ignore_joints else []
        with suppress_stdout_stderr():
            # Since parsing URDF causes a lot of warning messages which can't be deactivated, we suppress them
            self.urdf_object = URDFObject(urdf_path)
        self.joint_types = {joint.name: joint.type for joint in self.urdf_object.joints}
        self.joint_actuators: Optional[Dict] = parse_mjcf_actuators(mjcf_path) if mjcf_path is not None else None
        self.kinematic_chains: Dict[str, KinematicChainDescription] = {}
        self.cameras: Dict[str, CameraDescription] = {}
        self.links: List[str] = [l.name for l in self.urdf_object.links]
        self.joints: List[str] = [j.name for j in self.urdf_object.joints]
        self.virtual_mobile_base_joints: Optional[VirtualMobileBaseJoints] = virtual_mobile_base_joints
        self.gripper_name = gripper_name
        self.neck = {}

    def add_arm(self, end_link: str,
                arm_type: Arms = Arms.RIGHT,
                arm_name: str = "manipulator",
                arm_home_values: Optional[Dict[str, float]] = None,
                arm_start: Optional[str] = None) -> KinematicChainDescription:
        """
        Creates and adds an arm to the RobotDescription.

        :param end_link: Last link of the arm
        :param arm_type: Type of the arm
        :param arm_name: Name of the arm
        :param arm_home_values: Dictionary of joint names and their home values (default configuration) (e.g. park arms)
        :param arm_start: Start link of the arm
        """
        if arm_start is None:
            arm_start = self.base_link if self.torso_link == '' else self.torso_link
        arm = KinematicChainDescription(arm_name, arm_start, end_link,
                                        self.urdf_object, arm_type=arm_type)

        if arm_home_values is not None:
            arm.add_static_joint_states(StaticJointState.Park, arm_home_values)

        self.add_kinematic_chain_description(arm)
        return arm

    @property
    def has_actuators(self):
        """
        Property to check if the robot has actuators defined in the MJCF file.

        :return: True if the robot has actuators, False otherwise
        """
        return self.joint_actuators is not None

    def get_actuator_for_joint(self, joint: str) -> Optional[str]:
        """
        Get the actuator name for a given joint.

        :param joint: Name of the joint
        :return: Name of the actuator
        """
        if self.has_actuators:
            return self.joint_actuators.get(joint)
        return None

    def add_kinematic_chain_description(self, chain: KinematicChainDescription):
        """
        Adds a KinematicChainDescription object to the RobotDescription. The chain is stored with the name of the chain
        as key.

        :param chain: KinematicChainDescription object to add
        """
        if chain.name in self.kinematic_chains.keys():
            raise ValueError(f"Chain {chain.name} already exists for robot {self.name}")
        self.kinematic_chains[chain.name] = chain

    def add_kinematic_chain(self, name: str, start_link: str, end_link: str):
        """
        Creates and adds a KinematicChainDescription object to the RobotDescription.

        :param name: Name of the KinematicChainDescription object
        :param start_link: First link of the chain
        :param end_link: Last link of the chain
        """
        if name in self.kinematic_chains.keys():
            raise ValueError(f"Chain {name} already exists for robot {self.name}")
        chain = KinematicChainDescription(name, start_link, end_link, self.urdf_object)
        self.add_kinematic_chain_description(chain)

    def add_camera_description(self, camera: CameraDescription):
        """
        Adds a CameraDescription object to the RobotDescription. The camera is stored with the name of the camera as key.
        :param camera: The CameraDescription object to add
        """
        name = camera.name
        if name in self.cameras.keys():
            raise ValueError(f"Camera {name} already exists for robot {self.name}")
        self.cameras[name] = camera

    def add_camera(self, name: str, camera_link: str, minimal_height: float, maximal_height: float):
        """
        Creates and adds a CameraDescription object to the RobotDescription. Minimal and maximal height of the camera are
        relevant if the robot has a moveable torso or the camera is mounted on a moveable part of the robot. Otherwise
        both values can be the same.

        :param name: Name of the CameraDescription object
        :param camera_link: Link of the camera in the URDF
        :param minimal_height: Minimal height of the camera
        :param maximal_height: Maximal height of the camera
        :return:
        """
        camera_desc = CameraDescription(name, camera_link, minimal_height, maximal_height)
        self.cameras[name] = camera_desc

    def get_manipulator_chains(self) -> List[KinematicChainDescription]:
        """
        Get a list of all manipulator chains of the robot which posses an end effector.

        :return: A list of KinematicChainDescription objects
        """
        result = []
        for chain in self.kinematic_chains.values():
            if chain.end_effector:
                result.append(chain)
        return result

    def get_camera_frame(self, robot_object_name: str = None) -> str:
        """
        Quick method to get the name of a link of a camera. Uses the first camera in the list of cameras.

        :return: A name of the link of a camera
        """
        if robot_object_name is None:
            return f"{self.name}/{self.get_camera_link()}"
        return f"{robot_object_name}/{self.get_camera_link()}"

    def get_camera_link(self) -> str:
        """
        Quick method to get the name of a link of a camera. Uses the first camera in the list of cameras.

        :return: A name of the link of a camera
        """
        return self.get_default_camera().link_name

    def get_default_camera(self) -> CameraDescription:
        """
        Get the first camera in the list of cameras.

        :return: A CameraDescription object
        """
        return self.cameras[list(self.cameras.keys())[0]]

    def get_static_joint_chain(self, kinematic_chain_name: str, configuration_name: Union[str, Enum]):
        """
        Get the static joint states of a kinematic chain for a specific configuration. When trying to access one of
        the robot arms the function `:func: get_arm_chain` should be used.

        :param kinematic_chain_name:
        :param configuration_name:
        :return:
        """
        if kinematic_chain_name in self.kinematic_chains.keys():
            if configuration_name in self.kinematic_chains[kinematic_chain_name].static_joint_states.keys():
                return self.kinematic_chains[kinematic_chain_name].static_joint_states[configuration_name]
            else:
                raise ValueError(
                    f"There is no static joint state with the name {configuration_name} for Kinematic chain {kinematic_chain_name} of robot {self.name}. "
                    f"The following keys are available: {list(self.kinematic_chains[kinematic_chain_name].static_joint_states.keys())}")
        else:
            raise ValueError(f"There is no KinematicChain with name {kinematic_chain_name} for robot {self.name}. "
                             f"The following chains are available: {list(self.kinematic_chains.keys())}")

    def get_offset(self, name: str) -> Optional[PoseStamped]:
        """
        Returns the offset of a Joint in the URDF.

        :param name: The name of the Joint for which the offset will be returned.
        :return: The offset of the Joint
        """
        if name not in self.urdf_object.joint_map.keys():
            logerr(f"The name: {name} is not part of this robot URDF")
            return PoseStamped()

        offset = self.urdf_object.joint_map[name].origin
        return offset if offset else PoseStamped()

    def get_parent(self, name: str) -> str:
        """
        Get the parent of a link or joint in the URDF. Always returns the imeadiate parent, for a link this is a joint
        and vice versa.

        :param name: Name of the link or joint in the URDF
        :return: Name of the parent link or joint
        """
        if name not in self.links and name not in self.joints:
            raise ValueError(f"Link or joint {name} not found in URDF")
        if name in self.links:
            if name in self.urdf_object.parent_map:
                parent_joint, parent_link = self.urdf_object.parent_map[name]
                return parent_joint
            else:
                raise ValueError(f"Link {name} has no parent")
        elif name in self.joints:
            parent_link = self.urdf_object.joint_map[name].parent
            return parent_link

    def get_child(self, name: str, return_multiple_children: bool = False) -> Union[str, List[str]]:
        """
        Get the child of a link or joint in the URDF. Always returns the immediate child, for a link this is a joint
        and vice versa. Since a link can have multiple children, the return_multiple_children parameter can be set to
        True to get a list of all children.

        :param name: Name of the link or joint in the URDF
        :param return_multiple_children: If True, a list of all children is returned
        :return: Name of the child link or joint or a list of all children
        """
        if name not in self.links and name not in self.joints:
            raise ValueError(f"Link or joint {name} not found in URDF")
        if name in self.links:
            if name in self.urdf_object.child_map:
                children = self.urdf_object.child_map[name]
                # A link can have multiple children
                child_joints = [child[0] for child in children]
                if return_multiple_children:
                    child_joints.reverse()
                    return child_joints
                else:
                    return child_joints[0]

            else:
                raise ValueError(f"Link {name} has no children")
        elif name in self.joints:
            child_link = self.urdf_object.joint_map[name].child
            return child_link

    def get_arm_tool_frame(self, arm: Arms) -> str:
        """
        Get the name of the tool frame of a specific arm.

        :param arm: Arm for which the tool frame should be returned
        :return: The name of the link of the tool frame in the URDF.
        """
        chain = self.get_arm_chain(arm)
        return chain.get_tool_frame()

    def get_arm_chain(self, arm: Arms) -> Union[KinematicChainDescription, List[KinematicChainDescription]]:
        """
        Get the kinematic chain of a specific arm. If the arm is set to BOTH, all kinematic chains are returned.

        :param arm: Arm for which the chain should be returned
        :return: KinematicChainDescription object of the arm
        """
        if arm == Arms.BOTH:
            return list(filter(lambda chain: chain.arm_type is not None, self.kinematic_chains.values()))
        for chain in self.kinematic_chains.values():
            if chain.arm_type == arm:
                return chain
        raise ValueError(f"There is no Kinematic Chain for the Arm {arm}")

    def set_neck(self, yaw_joint: Optional[str] = None, pitch_joint: Optional[str] = None,
                 roll_joint: Optional[str] = None):
        """
        Defines the neck configuration of the robot by setting the yaw, pitch, and roll
        joints along with their corresponding links.

        :param yaw_joint: The joint name for the yaw movement of the neck.
        :param pitch_joint: The joint name for the pitch movement of the neck.
        :param roll_joint: The joint name for the roll movement of the neck.
        """
        if yaw_joint:
            self.neck["yaw"] = [self.get_child(yaw_joint), yaw_joint]
        if pitch_joint:
            self.neck["pitch"] = [self.get_child(pitch_joint), pitch_joint]
        if roll_joint:
            self.neck["roll"] = [self.get_child(roll_joint), roll_joint]

    def get_neck(self) -> Dict[str, List[Optional[str]]]:
        """
        Retrieves the neck configuration of the robot, including links and joints for yaw,
        pitch, and roll.

        :return: A dictionary containing the neck configuration. Keys are yaw, pitch, and roll. Values are [link, joint].
        """
        return self.neck

    def load(self):
        """
        Loads the robot description in the robot description manager, can be overridden to take more parameter into
        account.
        """
        RobotDescription.current_robot_description = self

    def unload(self):
        """
        Unloads the robot description in the robot description manager, can be overridden to take more parameter into
        account.
        """
        RobotDescription.current_robot_description = None


class KinematicChainDescription:
    """
    Represents a kinematic chain of a robot. A Kinematic chain is a chain of links and joints that are connected to each
    other and can be moved.

    This class contains all necessary information about the chain, like the start and end
    link, the URDF object and the joints of the chain.
    """

    name: str
    """
    Name of the chain
    """
    start_link: str
    """
    First link of the chain
    """
    end_link: str
    """
    Last link of the chain
    """
    urdf_object: URDFObject
    """
    Parsed URDF of the robot
    """
    include_fixed_joints: bool
    """
    If True, fixed joints are included in the chain
    """
    link_names: List[str]
    """
    List of all links in the chain
    """
    joint_names: List[str]
    """
    List of all joints in the chain
    """
    end_effector: EndEffectorDescription
    """
    End effector of the chain, if there is one
    """
    arm_type: Arms
    """
    Type of the arm, if the chain is an arm
    """
    static_joint_states: Dict[Union[str, Enum], Dict[str, float]]
    """
    Dictionary of static joint states for the chain
    """

    def __init__(self, name: str, start_link: str, end_link: str, urdf_object: URDFObject, arm_type: Arms = None,
                 include_fixed_joints=False):
        """
        Initialize the KinematicChainDescription object.

        :param name: Name of the chain
        :param start_link: First link of the chain
        :param end_link: Last link of the chain
        :param urdf_object: URDF object of the robot which is used to get the chain
        :param arm_type: Type of the arm, if the chain is an arm
        :param include_fixed_joints: If True, fixed joints are included in the chain
        """
        self.name: str = name
        self.start_link: str = start_link
        self.end_link: str = end_link
        self.urdf_object: URDFObject = urdf_object
        self.include_fixed_joints: bool = include_fixed_joints
        self.link_names: List[str] = []
        self.joint_names: List[str] = []
        self.arm_type: Arms = arm_type
        self.static_joint_states: Dict[Union[str, Enum], Dict[str, float]] = {}
        self.end_effector = None

        self._init_links()
        self._init_joints()

    def _init_links(self):
        """
        Initializes the links of the chain by getting the chain from the URDF object.
        """
        self.link_names = self.urdf_object.get_chain(self.start_link, self.end_link, joints=False)

    def _init_joints(self):
        """
        Initializes the joints of the chain by getting the chain from the URDF object.
        """
        joints = self.urdf_object.get_chain(self.start_link, self.end_link, links=False)
        self.joint_names = list(filter(lambda j: self.urdf_object.joint_map[j].type != JointType.FIXED
                                                 or self.include_fixed_joints, joints))

    def create_end_effector(self,
                            name: str,
                            tool_frame,
                            opened_joint_values: Dict[str, float],
                            closed_joint_values: Dict[str, float],
                            relative_dir: Optional[str] = None,
                            resources_dir: Optional[str] = None,
                            description_name: str = "gripper",
                            opening_distance: Optional[float] = None) -> EndEffectorDescription:
        """
        Create a gripper end effector description.

        :param name: The name of the gripper.
        :param tool_frame: The name of the tool frame.
        :param opened_joint_values: The joint values when the gripper is open.
        :param closed_joint_values: The joint values when the gripper is closed.
        :param relative_dir: The relative directory of the gripper in the Multiverse resources/robots directory.
        :param resources_dir: The path to the resources directory.
        :param description_name: The name of the gripper description.
        :param opening_distance: The openning distance of the gripper.
        :return: The gripper end effector description.
        """
        if resources_dir is None:
            resources_dir = find_multiverse_resources_path()
        if relative_dir is not None:
            gripper_filename = get_robot_description_path(relative_dir, name,
                                                          description_type=DescriptionType.URDF,
                                                          resources_dir=resources_dir)
            gripper_urdf_obj = URDFObject(gripper_filename)
            gripper_object_name = name
        else:
            gripper_urdf_obj = self.urdf_object
            gripper_object_name = None
        gripper = EndEffectorDescription(description_name, name, tool_frame,
                                         gripper_urdf_obj, gripper_object_name=gripper_object_name,
                                         opening_distance=opening_distance)

        gripper.add_static_joint_states(GripperState.OPEN, opened_joint_values)
        gripper.add_static_joint_states(GripperState.CLOSE, closed_joint_values)
        self.end_effector = gripper
        return gripper

    def get_joints(self) -> List[str]:
        """
        Get a list of all joints of the chain.

        :return: List of joint names
        """
        return self.joint_names

    def get_links(self) -> List[str]:
        """
        :return: A list of all links of the chain.
        """
        return self.link_names

    @property
    def links(self) -> List[str]:
        """
        Property to get the links of the chain.

        :return: List of link names
        """
        return self.get_links()

    @property
    def joints(self) -> List[str]:
        """
        Property to get the joints of the chain.

        :return: List of joint names
        """
        return self.get_joints()

    def add_static_joint_states(self, name: Union[str, Enum], states: dict):
        """
        Adds static joint states to the chain. These define a specific configuration of the chain.

        :param name: Name of the static joint states
        :param states: Dictionary of joint names and their values
        """
        for joint_name, value in states.items():
            if joint_name not in self.joint_names:
                raise ValueError(f"Joint {joint_name} is not part of the chain")
        self.static_joint_states[name] = states

    def get_static_joint_states(self, name: Union[str, Enum]) -> Dict[str, float]:
        """
        Get the dictionary of static joint states for a given name of the static joint states.

        :param name: Name of the static joint states
        :return: Dictionary of joint names and their values
        """
        try:
            return self.static_joint_states[name]
        except KeyError:
            logerr(f"Static joint states for chain {name} not found")

    def get_tool_frame(self) -> str:
        """
        Get the name of the tool frame of the end effector of this chain, if it has an end effector.

        :return: The name of the link of the tool frame in the URDF.
        """
        if self.end_effector:
            return self.end_effector.tool_frame
        else:
            raise ValueError(f"The Kinematic chain {self.name} has no end-effector")

    def get_static_gripper_state(self, state: GripperState) -> Dict[str, float]:
        """
        Get the static joint states for the gripper of the chain.

        :param state: Name of the static joint states
        :return: Dictionary of joint names and their values
        """
        if self.end_effector:
            return self.end_effector.static_joint_states[state]
        else:
            raise ValueError(f"The Kinematic chain {self.name} has no end-effector")


class CameraDescription:
    """
    Represents a camera mounted on a robot. Contains all necessary information about the camera, like the link name,
    minimal and maximal height, horizontal and vertical angle and the front facing axis.
    """
    name: str
    """
    Name of the camera
    """
    link_name: str
    """
    Name of the link in the URDF
    """
    minimal_height: float
    """
    Minimal height the camera can be at
    """
    maximal_height: float
    """
    Maximal height the camera can be at
    """
    horizontal_angle: float
    """
    Horizontal opening angle of the camera
    """
    vertical_angle: float
    """
    Vertical opening angle of the camera
    """
    front_facing_axis: List[int]
    """
    Axis along which the camera is taking the image
    """

    def __init__(self, name: str, link_name: str, minimal_height: float, maximal_height: float,
                 horizontal_angle: float = 20, vertical_angle: float = 20, front_facing_axis: List[float] = None):
        """
        Initialize the CameraDescription object.

        :param name: Name of the camera
        :param link_name: Name of the link in the URDF
        :param minimal_height: Minimal height the camera can be at
        :param maximal_height: Maximal height the camera can be at
        :param horizontal_angle: Horizontal opening angle of the camera
        :param vertical_angle: Vertical opening angle of the camera
        :param front_facing_axis: Axis along which the camera taking the image
        """
        self.name: str = name
        self.link_name: str = link_name
        self.minimal_height: float = minimal_height
        self.maximal_height: float = maximal_height
        self.horizontal_angle: float = horizontal_angle
        self.vertical_angle: float = vertical_angle
        self.front_facing_axis: List[int] = front_facing_axis if front_facing_axis else [0, 0, 1]


class EndEffectorDescription:
    """
    Describes an end effector of robot. Contains all necessary information about the end effector, like the
    base link, the tool frame, the URDF object and the static joint states.
    """
    name: str
    """
    Name of the end effector
    """
    start_link: str
    """
    Root link of the end effector, every link below this link in the URDF is part of the end effector
    """
    tool_frame: str
    """
    Name of the tool frame link in the URDf
    """
    urdf_object: URDFObject
    """
    Parsed URDF of the robot
    """
    link_names: List[str]
    """
    List of all links in the end effector
    """
    joint_names: List[str]
    """
    List of all joints in the end effector
    """
    static_joint_states: Dict[GripperState, Dict[str, float]]
    """
    Dictionary of static joint states for the end effector
    """
    end_effector_type: GripperType
    """
    Type of the gripper
    """
    opening_distance: float
    """
    Distance the gripper can open, in cm
    """
    gripper_object_name: Optional[str] = None
    """
    Name of the gripper of the robot if it has one, this is used when the gripper is a different Object with its own
    description file outside the robot description file.
    """
    fingers_link_names: Optional[List[str]] = None
    """
    List of all links of the fingers of the gripper
    """
    grasps: Dict[GraspDescription, List[float]]
    """
    Dictionary of all grasp orientations of the end effector
    """
    approach_axis: List[float]
    """
    Relative axis along which the end effector is approaching an object
    """

    def __init__(self, name: str, start_link: str, tool_frame: str, urdf_object: URDFObject,
                 gripper_object_name: Optional[str] = None, opening_distance: Optional[float] = None,
                 fingers_link_names: Optional[List[str]] = None):
        """
        Initialize the EndEffectorDescription object.

        :param name: Name of the end effector
        :param start_link: Root link of the end effector, every link below this link in the URDF is part of the end effector
        :param tool_frame: Name of the tool frame link in the URDf
        :param urdf_object: URDF object of the robot
        :param gripper_object_name: Name of the gripper if it is a separate Object outside the robot description.
        :param opening_distance: Distance the gripper can open, in cm.
        :param fingers_link_names: List of all link names of the fingers of the gripper if it has fingers.
        """
        self.name: str = name
        self.start_link: str = start_link
        self.tool_frame: str = tool_frame
        self.urdf_object: URDFObject = urdf_object
        self.link_names: List[str] = []
        self.joint_names: List[str] = []
        self.static_joint_states: Dict[GripperState, Dict[str, float]] = {}
        self._init_links_joints()
        self.gripper_object_name = gripper_object_name
        self.opening_distance: Optional[float] = opening_distance
        self.fingers_link_names: Optional[List[str]] = fingers_link_names
        self.grasps: Dict[GraspDescription, List[float]] = {}
        self.approach_axis = None

    def _init_links_joints(self):
        """
        Traverses the URDF object to get all links and joints of the end effector below the start link.1
        """
        start_link_obj = self.urdf_object.link_map[self.start_link]
        links = [start_link_obj.name]
        while len(links) != 0:
            link = links.pop()
            self.link_names.append(link)
            if link not in self.urdf_object.child_map:
                continue
            else:
                children = self.urdf_object.child_map[link]
            for joint, link in children:
                self.joint_names.append(joint)
                links.insert(0, link)

    def add_static_joint_states(self, name: GripperState, states: dict):
        """
        Adds static joint states to the end effector. These define a specific configuration of the end effector. Like
        open and close configurations of a gripper.

        :param name: Name of the static joint states
        :param states: Dictionary of joint names and their values
        """
        for joint_name, value in states.items():
            if joint_name not in self.joint_names:
                raise ValueError(f"Joint {joint_name} is not part of the chain")
        self.static_joint_states[name] = states

    @property
    def links(self) -> List[str]:
        """
        Property to get the links of the chain.

        :return: List of link names
        """
        return self.link_names

    @property
    def joints(self) -> List[str]:
        """
        Property to get the joints of the chain.

        :return: List of joint names
        """
        return self.joint_names

    def update_all_grasp_orientations(self, front_orientation: List[float]):
        """
        Generates all grasp quaternion orientations based on a given front-facing quaternion orientation in-place,
        covering combinations of side grasps (front, back, left, right),
        top/bottom grasps, and horizontal rotation options.

        :param front_orientation: A quaternion representing the front-facing orientation as [x, y, z, w] quaternion.
        """

        grasp_descriptions = [
            GraspDescription(side, vertical, horizontal)
            for side, vertical, horizontal in product(
                Rotations.SIDE_ROTATIONS.keys(),
                Rotations.VERTICAL_ROTATIONS.keys(),
                Rotations.HORIZONTAL_ROTATIONS.keys()
            )
        ]

        for grasp_description in grasp_descriptions:
            rotation = Rotations.SIDE_ROTATIONS[grasp_description.approach_direction]
            rotation = quaternion_multiply(rotation, Rotations.VERTICAL_ROTATIONS[grasp_description.vertical_alignment])
            rotation = quaternion_multiply(rotation, Rotations.HORIZONTAL_ROTATIONS[grasp_description.rotate_gripper])

            orientation = quaternion_multiply(rotation, front_orientation)

            norm = math.sqrt(sum(comp ** 2 for comp in orientation))
            orientation = [comp / norm for comp in orientation]

            self.grasps[grasp_description] = orientation

    def get_grasp(self, approach_direction: Grasp, vertical_alignment: Grasp = None, rotate_gripper: bool = False) -> List[float]:
        """
        Retrieves the quaternion orientation of the end effector for a specific grasp.

        :param approach_direction: The approach direction of the end effector.
        :param vertical_alignment: The vertical alignment of the end effector.
        :param rotate_gripper: True, the gripper should be rotated 90°.

        :return: List of floats representing the quaternion orientation of the end effector
        """
        grasp_description = GraspDescription(approach_direction, vertical_alignment, rotate_gripper)
        return self.grasps[grasp_description]

    def set_approach_axis(self, axis: List[float]):
        """
        Sets the approach axis for the robot's palm.

        :param axis: A list representing the approach axis.
        """
        self.approach_axis = axis

    def get_approach_axis(self) -> List[float]:
        """
        Retrieves the approach axis.

        :return: A list representing the approach axis.
        """
        if self.approach_axis is None:
            front_grasp = self.get_grasp(Grasp.FRONT, None, False)
            self.approach_axis = R.from_quat(front_grasp).as_matrix()[0]
        return self.approach_axis


def create_manipulator_description(data: ManipulatorData,
                                   urdf_filename: str,
                                   mjcf_filename: Optional[str] = None) -> RobotDescription:
    """
    Create a robot description from a ManipulatorData object.

    :param data: ManipulatorData object containing all necessary information about the manipulator.
    :param urdf_filename: Path to the URDF file of the robot.
    :param mjcf_filename: Path to the MJCF file of the robot.
    :return: A RobotDescription object
    """
    gripper_object_name = None if data.gripper_relative_dir is None else data.gripper_name
    robot_description = RobotDescription(data.name, data.base_link, "", "",
                                         urdf_filename, mjcf_path=mjcf_filename, gripper_name=gripper_object_name)

    arm = robot_description.add_arm(data.arm_end_link,
                                    arm_home_values=dict(zip(data.joint_names, data.home_joint_values)))

    arm.create_end_effector(data.gripper_name, data.gripper_tool_frame,
                            dict(zip(data.gripper_joint_names, data.open_joint_values)),
                            dict(zip(data.gripper_joint_names, data.closed_joint_values)),
                            relative_dir=data.gripper_relative_dir, opening_distance=data.opening_distance)

    return robot_description



import traceback
import sqlalchemy
try:
    from .ormatic_interface import mapper_registry
except ImportError:
    pass
from ..designators.object_designator import *
import json

from ..ros import  loginfo, logwarn



def write_database_to_file(in_sessionmaker: sqlalchemy.orm.sessionmaker, filename: str,
                           b_write_to_console: bool = False):
    """
    Writes all Tables stored within the given session into a local file. File will be written in JSON Format

    :param in_sessionmaker: sessionmaker that allows us to access the Database
    :param filename: Filename of the logfile
    :param b_write_to_console: enables writing to the console. Default false
    """
    with in_sessionmaker() as session:
        with open("whatever.txt", "w") as f:
            to_json_dict = dict()
            for table in mapper_registry.metadata.sorted_tables:
                list_of_row = list()
                for column_object in session.query(table).all():
                    list_of_row.append(column_object)
                to_json_dict[table.name] = list_of_row
            json_data_dict = json.dumps(to_json_dict, default=str)
            f.write(json_data_dict)


def print_database(in_sessionmaker: sqlalchemy.orm.sessionmaker):
    """
    Prints all ORM Class data within the given Session.

    :param in_sessionmaker: Database Session which should be printed
    """
    with in_sessionmaker() as session:
        for table in mapper_registry.metadata.sorted_tables:
            try:
                smt = sqlalchemy.select('*').select_from(table)
                result = session.execute(smt).all()
                loginfo("Table: {}\tcontent:{}".format(table, result))
            except sqlalchemy.exc.ArgumentError as e:
                logwarn(e)


def update_primary_key(source_session_maker: sqlalchemy.orm.sessionmaker,
                       destination_session_maker: sqlalchemy.orm.sessionmaker):
    """
    Updates all the primary keys of the database associated with the destination engine, so that there will be no
    problems when merging it into the source database. In order to achieve this the highest id value of the source
    engine is searched and the primary keys of the destination database will get all the values following that.
    Cascading triggers in the database will take care of the rest. Careful 2023 this will not work in
    memory databases as there are no triggers.

    :param source_session_maker: Session maker of the source data_base
    :param destination_session_maker: Session maker of the destination data_base
    """
    destination_session = destination_session_maker()
    source_session = source_session_maker()
    sortedTables = mapper_registry.metadata.sorted_tables
    for table in sortedTables:
        try:
            list_of_primary_keys_of_this_table = table.primary_key.columns.values()
            for key in list_of_primary_keys_of_this_table:
                all_source_key_values = []
                all_destination_key_values = []
                for key_value_row in source_session.query(key).all():
                    all_source_key_values.append(key_value_row[0])  # get all values of key from source session
                for key_value_row in destination_session.query(key).all():
                    all_destination_key_values.append(key_value_row[0])  # get all values of key from source session

                highest_free_key_value = max(max(all_source_key_values, default=0),
                                             max(all_destination_key_values, default=0)) + 1
                results = destination_session.execute(sqlalchemy.select(table))
                for column_object in results:  # iterate over all columns
                    if column_object.__getattr__(key.name) in all_source_key_values:
                        loginfo(
                            "Found primary_key collision in table {} value: {} max value in memory {}".format(table,
                                                                                                              column_object.__getattr__(
                                                                                                                  key.name),
                                                                                                              highest_free_key_value))
                        mini_dict = dict()
                        mini_dict[key.name] = highest_free_key_value
                        update_statement = sqlalchemy.update(table).where(
                            table.c.id == column_object.__getattr__(key)).values(mini_dict)
                        destination_session.execute(update_statement)
                        highest_free_key_value += 1
            destination_session.commit()  # commit after every table
        except AttributeError as e:
            logwarn("Possible found abstract ORM class {}".format(e.__name__))
            logwarn(e)
    destination_session.close()


def copy_database(source_session_maker: sqlalchemy.orm.sessionmaker,
                  destination_session_maker: sqlalchemy.orm.sessionmaker):
    """
    Iterates through all tables within tht source database and merges them into the destination database. Careful
    this function does not check if there are any primary key collisions or updates any data.

     .. note::
        Ignores all previously detached data, could result in loss of information. During testing database objects
        sometimes had a detached twin. As a possible feature in the future it maybe useful to give the user an
        opportunity to decide what happens with the detached objects. Careful this could lead to duplicated data in the
        destination database.

    :param source_session_maker: Sessionmaker of the source database
    :param destination_session_maker: Sessionmaker of the destination database
    """

    with source_session_maker() as source_session, destination_session_maker() as destination_session:
        sorted_tables = mapper_registry.metadata.sorted_tables
        for table in sorted_tables:
            for value in source_session.query(table).all():
                insert_statement = sqlalchemy.insert(table).values(value)
                destination_session.execute(insert_statement)
            destination_session.commit()  # commit after every table


def update_primary_key_constrains(session_maker: sqlalchemy.orm.sessionmaker):
    """
    Iterates through all tables related to any ORM Class and sets in their corresponding foreign keys in the given
    endpoint to "ON UPDATE CASCADING".

        .. note::
            Careful currently only works on postgres databases.

    :param session_maker:
    :return: empty
    """
    with session_maker() as session:
        for table in mapper_registry.metadata.sorted_tables:
            try:
                foreign_key_statement = sqlalchemy.text(
                    "SELECT con.oid, con.conname, con.contype, con.confupdtype, con.confdeltype, con.confmatchtype, pg_get_constraintdef(con.oid) FROM pg_catalog.pg_constraint con INNER JOIN pg_catalog.pg_class rel ON rel.oid = con.conrelid INNER JOIN pg_catalog.pg_namespace nsp ON nsp.oid = connamespace WHERE rel.relname = '{}';".format(
                        table))
                response = session.execute(foreign_key_statement)
                loginfo(25 * '~' + "{}".format(table) + 25 * '~')
                for line in response:
                    if line.conname.endswith("fkey"):
                        if 'a' in line.confupdtype:  # a --> no action | if there is no action we set it to cascading
                            # Assumes there aren't any other constraints
                            drop_statement = sqlalchemy.text(
                                "alter table \"{}\" drop constraint \"{}\";".format(table,
                                                                                    line.conname))
                            drop_response = session.execute(
                                drop_statement)  # There is no real data coming back for this
                            alter_statement = sqlalchemy.text(
                                "alter table \"{}\" add constraint {} {} on update cascade;".format(
                                    table,
                                    line.conname,
                                    line.pg_get_constraintdef))
                            alter_response = session.execute(
                                alter_statement)  # There is no real data coming back for this
                            session.commit()
            except AttributeError:
                loginfo("Attribute Error: {} has no attribute __tablename__".format(table))


def migrate_neems(source_session_maker: sqlalchemy.orm.sessionmaker,
                  destination_session_maker: sqlalchemy.orm.sessionmaker):
    """
        Merges the database connected to the source session maker into the database connected to the destination session
        maker. Will first update the primary constrains inside the destination database (if needed). Afterwards
        updates the primary keys within the destination database (as there are cascading updates now) and then merges
        the source database into the destination.

         .. note::
            Assumes the destination database is a postgres database

        :param source_session_maker: Sessionmaker of the source database
        :param destination_session_maker: Sessionmaker of the destination database
        """

    update_primary_key_constrains(destination_session_maker)
    update_primary_key(source_session_maker, destination_session_maker)
    copy_database(source_session_maker, destination_session_maker)



# used for delayed evaluation of typing until python 3.11 becomes mainstream
from __future__ import annotations

import os
import threading
import time

import numpy as np
import pycram_bullet as p
import yaml
from typing_extensions import List, Optional, Dict, Any, Callable

from pycrap.ontologies import Floor
from ..datastructures.dataclasses import Color, AxisAlignedBoundingBox, MultiBody, VisualShape, BoxVisualShape, \
    ClosestPoint, LateralFriction, ContactPoint, ContactPointsList, ClosestPointsList, RayResult
from ..datastructures.enums import ObjectType, WorldMode, JointType
from ..datastructures.pose import PoseStamped, Point
from ..datastructures.world import World
from ..datastructures.world_entity import PhysicalBody
from ..object_descriptors.generic import ObjectDescription as GenericObjectDescription
from ..object_descriptors.urdf import ObjectDescription
from ..ros import  logwarn, loginfo
from ..validation.goal_validator import (validate_multiple_joint_positions, validate_joint_position,
                                         validate_object_pose, validate_multiple_object_poses)
from ..world_concepts.constraints import Constraint
from ..world_concepts.world_object import Object
from ..config.world_conf import WorldConfig

Link = ObjectDescription.Link
RootLink = ObjectDescription.RootLink
Joint = ObjectDescription.Joint


class BulletWorld(World):
    """
    This class represents a BulletWorld, which is a simulation environment that uses the Bullet Physics Engine. This
    class is the main interface to the Bullet Physics Engine and should be used to spawn Objects, simulate Physic and
    manipulate the Bullet World.
    """

    def __init__(self, mode: WorldMode = WorldMode.DIRECT, is_prospection: bool = False,
                 use_multiverse_for_real_world_simulation: bool = False):
        """
        Creates a new simulation, the type decides of the simulation should be a rendered window or just run in the
        background. There can only be one rendered simulation.
        The BulletWorld object also initializes the Events for attachment, detachment and for manipulating the world.

        :param mode: Can either be "GUI" for rendered window or "DIRECT" for non-rendered. The default is "GUI"
        :param is_prospection: For internal usage, decides if this BulletWorld should be used as a shadow world.
        :param use_multiverse_for_real_world_simulation: Whether to use the Multiverse for real world simulation.
        """
        super().__init__(mode=mode, is_prospection=is_prospection)

        if use_multiverse_for_real_world_simulation:
            self.add_multiverse_resources()

        # This disables file caching from PyBullet, since this would also cache
        # files that can not be loaded
        p.setPhysicsEngineParameter(enableFileCaching=0)

        # Needed to let the other thread start the simulation, before Objects are spawned.
        time.sleep(0.1)
        self.vis_axis: List[int] = []

        # Some default settings
        self.set_gravity([0, 0, -9.8])

        if not is_prospection:
            _ = Object("floor", Floor, "plane.urdf",
                       world=self)

    def add_multiverse_resources(self):
        """
        Add the Multiverse resources to the start of the data directories of the BulletWorld such they are searched
        first for the resources because the pycram objects need to have the same description as the Multiverse objects.
        """
        try:
            from .multiverse import Multiverse
            Multiverse.make_sure_multiverse_resources_are_added(self.conf.clear_cache_at_start)
        except ImportError:
            logwarn("Multiverse is not installed, please install it to use the Multiverse.")

    def _init_world(self, mode: WorldMode):
        self._gui_thread: Gui = Gui(self, mode)
        self._gui_thread.start()
        time.sleep(0.1)

    def load_generic_object_and_get_id(self, description: GenericObjectDescription,
                                       pose: Optional[PoseStamped] = None) -> int:
        """
        Creates a visual and collision box in the simulation.
        """
        # Create visual shape
        vis_shape = p.createVisualShape(p.GEOM_BOX, halfExtents=description.shape_data,
                                        rgbaColor=description.color.get_rgba(), physicsClientId=self.id)

        # Create collision shape
        col_shape = p.createCollisionShape(p.GEOM_BOX, halfExtents=description.shape_data, physicsClientId=self.id)

        # Create MultiBody with both visual and collision shapes
        obj_id = p.createMultiBody(baseMass=1.0, baseCollisionShapeIndex=col_shape, baseVisualShapeIndex=vis_shape,
                                   basePosition=description.origin.position.to_list(),
                                   baseOrientation=description.origin.orientation.to_list(), physicsClientId=self.id)

        if pose is not None:
            self._set_object_pose_by_id(obj_id, pose)
        # Assuming you have a list to keep track of created objects
        return obj_id

    def load_object_and_get_id(self, path: Optional[str] = None, pose: Optional[PoseStamped] = None,
                               obj_type: Optional[ObjectType] = None) -> int:
        if pose is None:
            pose = PoseStamped()
        return self._load_object_and_get_id(path, pose)

    def _load_object_and_get_id(self, path: str, pose: PoseStamped) -> int:
        if path is None:
            raise ValueError("Path to the object file is required.")
        return p.loadURDF(path,
                          basePosition=pose.position.to_list(),
                          baseOrientation=pose.orientation.to_list(), physicsClientId=self.id)

    def _remove_visual_object(self, obj_id: int) -> bool:
        self._remove_body(obj_id)
        return True

    def remove_object_from_simulator(self, obj: Object) -> bool:
        self._remove_body(obj.id)
        return True

    def _remove_body(self, body_id: int) -> Any:
        """
        Remove a body from PyBullet using the body id.

        :param body_id: The id of the body.
        """
        return p.removeBody(body_id, self.id)

    def add_constraint(self, constraint: Constraint) -> int:

        constraint_id = p.createConstraint(constraint.parent_object_id,
                                           constraint.parent_link_id,
                                           constraint.child_object_id,
                                           constraint.child_link_id,
                                           constraint.type.value,
                                           constraint.axis_as_list,
                                           constraint.position_wrt_parent_as_list,
                                           constraint.position_wrt_child_as_list,
                                           constraint.orientation_wrt_parent_as_list,
                                           constraint.orientation_wrt_child_as_list,
                                           physicsClientId=self.id)
        return constraint_id

    def remove_constraint(self, constraint_id):
        p.removeConstraint(constraint_id, physicsClientId=self.id)

    def _get_joint_position(self, joint: ObjectDescription.Joint) -> float:
        return p.getJointState(joint.object_id, joint.id, physicsClientId=self.id)[0]

    def get_object_joint_names(self, obj: Object) -> List[str]:
        return [p.getJointInfo(obj.id, i, physicsClientId=self.id)[1].decode('utf-8')
                for i in range(self.get_object_number_of_joints(obj))]

    def get_multiple_link_poses(self, links: List[Link]) -> Dict[str, PoseStamped]:
        return {link.name: self.get_link_pose(link) for link in links}

    def get_multiple_link_positions(self, links: List[Link]) -> Dict[str, List[float]]:
        return {link.name: self.get_link_position(link) for link in links}

    def get_multiple_link_orientations(self, links: List[Link]) -> Dict[str, List[float]]:
        return {link.name: self.get_link_orientation(link) for link in links}

    def get_link_position(self, link: Link) -> List[float]:
        return self.get_link_pose(link).position.to_list()

    def get_link_orientation(self, link: Link) -> List[float]:
        return self.get_link_pose(link).orientation.to_list()

    def get_link_pose(self, link: ObjectDescription.Link) -> PoseStamped:
        bullet_link_state = p.getLinkState(link.object_id, link.id, physicsClientId=self.id)
        return PoseStamped.from_list(*bullet_link_state[4:6])

    def get_object_link_names(self, obj: Object) -> List[str]:
        num_links = self.get_object_number_of_links(obj)
        return [p.getJointInfo(obj.id, i, physicsClientId=self.id)[12].decode('utf-8')
                for i in range(num_links)]

    def get_object_number_of_links(self, obj: Object) -> int:
        return p.getNumJoints(obj.id, physicsClientId=self.id)

    get_object_number_of_joints = get_object_number_of_links

    def perform_collision_detection(self) -> None:
        p.performCollisionDetection(physicsClientId=self.id)

    def get_body_contact_points(self, body: PhysicalBody) -> ContactPointsList:
        self.perform_collision_detection()
        body_data = self.get_body_and_link_id(body, index='A')
        points_list = p.getContactPoints(**body_data, physicsClientId=self.id)
        return ContactPointsList([ContactPoint(**self.parse_points_list_to_args(point)) for point in points_list
                                  if len(point) > 0])

    def get_contact_points_between_two_bodies(self, obj_a: PhysicalBody, obj_b: PhysicalBody) -> ContactPointsList:
        self.perform_collision_detection()
        body_a_data = self.get_body_and_link_id(obj_a, index='A')
        body_b_data = self.get_body_and_link_id(obj_b, index='B')
        points_list = p.getContactPoints(**body_a_data, **body_b_data, physicsClientId=self.id)
        return ContactPointsList([ContactPoint(**self.parse_points_list_to_args(point)) for point in points_list
                                  if len(point) > 0])

    def get_body_closest_points(self, body: PhysicalBody, max_distance: float) -> ClosestPointsList:
        all_obj_closest_points = [self.get_closest_points_between_two_bodies(body, other_body, max_distance)
                                  for other_body in self.objects if other_body != body]
        return ClosestPointsList([point for closest_points in all_obj_closest_points for point in closest_points])

    def get_closest_points_between_two_bodies(self, obj_a: PhysicalBody, obj_b: PhysicalBody,
                                              max_distance: float) -> ClosestPointsList:
        body_a_data = self.get_body_and_link_id(obj_a, index='A')
        body_b_data = self.get_body_and_link_id(obj_b, index='B')
        points_list = p.getClosestPoints(**body_a_data, **body_b_data, distance=max_distance, physicsClientId=self.id)
        return ClosestPointsList([ClosestPoint(**self.parse_points_list_to_args(point)) for point in points_list
                                  if len(point) > 0])

    @staticmethod
    def get_body_and_link_id(body: PhysicalBody, index='') -> Dict[str, int]:
        body_id, link_id = (body.object_id, body.id) if isinstance(body, Link) else (body.id, None)
        values_dict = {f"body{index}": body_id}
        if link_id is not None:
            values_dict.update({f"linkIndex{index}": link_id})
        return values_dict

    def parse_points_list_to_args(self, point: List) -> Dict:
        """
        Parses the list of points to a list of dictionaries with the keys as the names of the arguments of the
        ContactPoint class.

        :param point: The list of points.
        """
        return {"body_a": self.get_object_by_id(point[1]).get_link_by_id(point[3]),
                "body_b": self.get_object_by_id(point[2]).get_link_by_id(point[4]),
                "position_on_body_a": point[5],
                "position_on_body_b": point[6],
                "normal_on_body_b": point[7],
                "distance": point[8],
                "normal_force": point[9],
                "lateral_friction_1": LateralFriction(point[10], point[11]),
                "lateral_friction_2": LateralFriction(point[12], point[13])}

    @validate_multiple_joint_positions
    def _set_multiple_joint_positions(self, joint_positions: Dict[Joint, float]) -> bool:
        for joint, joint_position in joint_positions.items():
            self.reset_joint_position(joint, joint_position)
        return True

    @validate_joint_position
    def _reset_joint_position(self, joint: Joint, joint_position: float) -> bool:
        p.resetJointState(joint.object_id, joint.id, joint_position, physicsClientId=self.id)
        return True

    def _get_multiple_joint_positions(self, joints: List[Joint]) -> Dict[str, float]:
        return {joint.name: self.get_joint_position(joint) for joint in joints}

    @validate_multiple_object_poses
    def reset_multiple_objects_base_poses(self, objects: Dict[Object, PoseStamped]) -> bool:
        for obj, pose in objects.items():
            self.reset_object_base_pose(obj, pose)
        return True

    @validate_object_pose
    def reset_object_base_pose(self, obj: Object, pose: PoseStamped) -> bool:
        return self._set_object_pose_by_id(obj.id, pose)

    def _set_object_pose_by_id(self, obj_id: int, pose: PoseStamped) -> bool:
        p.resetBasePositionAndOrientation(obj_id, pose.position.to_list(), pose.orientation.to_list(),
                                          physicsClientId=self.id)
        return True

    def step(self, func: Optional[Callable[[], None]] = None, step_seconds: Optional[float] = None) -> None:
        if func is not None:
            logwarn("The BulletWorld step function does not support a function argument.")
        if step_seconds is not None:
            logwarn("The BulletWorld step function does not support a step_seconds argument.")
        p.stepSimulation(physicsClientId=self.id)

    def get_multiple_object_poses(self, objects: List[Object]) -> Dict[str, PoseStamped]:
        return {obj.name: self.get_object_pose(obj) for obj in objects}

    def get_multiple_object_positions(self, objects: List[Object]) -> Dict[str, List[float]]:
        return {obj.name: self.get_object_pose(obj).position.to_list() for obj in objects}

    def get_multiple_object_orientations(self, objects: List[Object]) -> Dict[str, List[float]]:
        return {obj.name: self.get_object_pose(obj).orientation.to_list() for obj in objects}

    def get_object_position(self, obj: Object) -> List[float]:
        return self.get_object_pose(obj).position.to_list()

    def get_object_orientation(self, obj: Object) -> List[float]:
        return self.get_object_pose(obj).orientation.to_list()

    def get_object_pose(self, obj: Object) -> PoseStamped:
        return PoseStamped.from_list(*p.getBasePositionAndOrientation(obj.id, physicsClientId=self.id), frame="map")

    def set_link_color(self, link: ObjectDescription.Link, rgba_color: Color):
        p.changeVisualShape(link.object_id, link.id, rgbaColor=rgba_color.get_rgba(), physicsClientId=self.id)

    def get_link_color(self, link: ObjectDescription.Link) -> Color:
        return self.get_colors_of_object_links(link.object)[link.name]

    def get_colors_of_object_links(self, obj: Object) -> Dict[str, Color]:
        visual_data = p.getVisualShapeData(obj.id, physicsClientId=self.id)
        link_id_to_name = {v: k for k, v in obj.link_name_to_id.items()}
        links = list(map(lambda x: link_id_to_name[x[1]], visual_data))
        colors = list(map(lambda x: Color.from_rgba(x[7]), visual_data))
        link_to_color = dict(zip(links, colors))
        return link_to_color

    def get_object_axis_aligned_bounding_box(self, obj: Object) -> AxisAlignedBoundingBox:
        return AxisAlignedBoundingBox.from_min_max(*p.getAABB(obj.id, physicsClientId=self.id))

    def get_link_axis_aligned_bounding_box(self, link: ObjectDescription.Link) -> AxisAlignedBoundingBox:
        return AxisAlignedBoundingBox.from_min_max(*p.getAABB(link.object_id, link.id, physicsClientId=self.id))

    def set_realtime(self, real_time: bool) -> None:
        p.setRealTimeSimulation(1 if real_time else 0, physicsClientId=self.id)

    def set_gravity(self, gravity_vector: List[float]) -> None:
        p.setGravity(gravity_vector[0], gravity_vector[1], gravity_vector[2], physicsClientId=self.id)

    def disconnect_from_physics_server(self):
        p.disconnect(physicsClientId=self.id)

    def join_threads(self):
        """
        Joins the GUI thread if it exists.
        """
        self.join_gui_thread_if_exists()

    def join_gui_thread_if_exists(self):
        if self._gui_thread:
            self._gui_thread.join()

    def save_physics_simulator_state(self, state_id: Optional[int] = None, use_same_id: bool = False) -> int:
        return p.saveState(physicsClientId=self.id)

    def restore_physics_simulator_state(self, state_id):
        p.restoreState(state_id, physicsClientId=self.id)

    def remove_physics_simulator_state(self, state_id: int):
        p.removeState(state_id, physicsClientId=self.id)

    def _add_vis_axis(self, pose: PoseStamped,
                      length: Optional[float] = 0.2) -> int:
        """
        Creates a Visual object which represents the coordinate frame at the given
        position and orientation. There can be an unlimited amount of vis axis objects.

        :param pose: The pose at which the axis should be spawned
        :param length: Optional parameter to configure the length of the axes
        :return: The id of the spawned object
        """

        pose_in_map = self.local_transformer.transform_pose(pose, "map")

        box_vis_shape = BoxVisualShape(Color(1, 0, 0, 0.8), [length, 0.01, 0.01], [length, 0.01, 0.01])
        vis_x = self.create_visual_shape(box_vis_shape)

        box_vis_shape = BoxVisualShape(Color(0, 1, 0, 0.8), [0.01, length, 0.01], [0.01, length, 0.01])
        vis_y = self.create_visual_shape(box_vis_shape)

        box_vis_shape = BoxVisualShape(Color(0, 0, 1, 0.8), [0.01, 0.01, length], [0.01, 0.01, length])
        vis_z = self.create_visual_shape(box_vis_shape)

        multibody = MultiBody(base_visual_shape_index=-1, base_pose=pose_in_map,
                              link_visual_shape_indices=[vis_x, vis_y, vis_z],
                              link_poses=[PoseStamped(), PoseStamped(), PoseStamped()], link_masses=[1.0, 1.0, 1.0],
                              link_inertial_frame_poses=[PoseStamped(), PoseStamped(), PoseStamped()], link_parent_indices=[0, 0, 0],
                              link_joint_types=[JointType.FIXED.value, JointType.FIXED.value, JointType.FIXED.value],
                              link_joint_axis=[Point(x=1.0, y=0.0, z=0.0), Point(x=0.0, y=1.0, z=0.0), Point(x=0.0, y=0.0, z=1.0)],
                              link_collision_shape_indices=[-1, -1, -1])

        body_id = self._create_multi_body(multibody)
        self.vis_axis.append(body_id)
        return body_id

    def _remove_vis_axis(self) -> None:
        """
        Removes all spawned vis axis objects that are currently in this BulletWorld.
        """
        for vis_id in self.vis_axis:
            p.removeBody(vis_id, physicsClientId=self.id)
        self.vis_axis = []

    def _ray_test(self, from_position: List[float], to_position: List[float]) -> RayResult:
        res = p.rayTest(from_position, to_position, physicsClientId=self.id)
        return RayResult(*res[0])

    def _ray_test_batch(self, from_positions: List[List[float]], to_positions: List[List[float]],
                        num_threads: int = 1) -> List[RayResult]:
        result = p.rayTestBatch(from_positions, to_positions, numThreads=num_threads,
                                physicsClientId=self.id)
        return [RayResult(*r) for r in result] if result else None

    def _create_visual_shape(self, visual_shape: VisualShape) -> int:
        return p.createVisualShape(visual_shape.visual_geometry_type.value,
                                   rgbaColor=visual_shape.rgba_color.get_rgba(),
                                   visualFramePosition=visual_shape.visual_frame_position,
                                   physicsClientId=self.id, **visual_shape.shape_data())

    def _create_multi_body(self, multi_body: MultiBody) -> int:
        return p.createMultiBody(baseVisualShapeIndex=-multi_body.base_visual_shape_index,
                                 linkVisualShapeIndices=multi_body.link_visual_shape_indices,
                                 basePosition=multi_body.base_pose.position.to_list(),
                                 baseOrientation=multi_body.base_pose.orientation.to_list(),
                                 linkPositions=[pose.position.to_list() for pose in multi_body.link_poses],
                                 linkMasses=multi_body.link_masses,
                                 linkOrientations=[pose.orientation.to_list() for pose in multi_body.link_poses],
                                 linkInertialFramePositions=[pose.position.to_list()
                                                             for pose in multi_body.link_inertial_frame_poses],
                                 linkInertialFrameOrientations=[pose.orientation.to_list()
                                                                for pose in multi_body.link_inertial_frame_poses],
                                 linkParentIndices=multi_body.link_parent_indices,
                                 linkJointTypes=multi_body.link_joint_types,
                                 linkJointAxis=[[point.x, point.y, point.z] for point in multi_body.link_joint_axis],
                                 linkCollisionShapeIndices=multi_body.link_collision_shape_indices)

    def get_images_for_target(self,
                              target_pose: PoseStamped,
                              cam_pose: PoseStamped,
                              size: Optional[int] = 256) -> List[np.ndarray]:
        # TODO: Might depend on robot cameras, if so please add these camera parameters to RobotDescription object
        # TODO: of your robot with a CameraDescription object.
        fov = 90
        aspect = size / size
        near = 0.2
        far = 100

        view_matrix = p.computeViewMatrix(cam_pose.position.to_list(), target_pose.position.to_list(), [0, 0, 1])
        projection_matrix = p.computeProjectionMatrixFOV(fov, aspect, near, far)
        return list(p.getCameraImage(size, size, view_matrix, projection_matrix,
                                     physicsClientId=self.id))[2:5]

    def _add_text(self, text: str, position: List[float], orientation: Optional[List[float]] = None,
                  size: Optional[float] = None, color: Optional[Color] = Color(), life_time: Optional[float] = 0,
                  parent_object_id: Optional[int] = None, parent_link_id: Optional[int] = None) -> int:
        args = {}
        if orientation:
            args["textOrientation"] = orientation
        if size:
            args["textSize"] = size
        if life_time:
            args["lifeTime"] = life_time
        if parent_object_id:
            args["parentObjectUniqueId"] = parent_object_id
        if parent_link_id:
            args["parentLinkIndex"] = parent_link_id
        return p.addUserDebugText(text, position, color.get_rgb(), physicsClientId=self.id, **args)

    def _remove_text(self, text_id: Optional[int] = None) -> None:
        if text_id is not None:
            p.removeUserDebugItem(text_id, physicsClientId=self.id)
        else:
            p.removeAllUserDebugItems(physicsClientId=self.id)

    def enable_joint_force_torque_sensor(self, obj: Object, fts_joint_idx: int) -> None:
        p.enableJointForceTorqueSensor(obj.id, fts_joint_idx, enableSensor=1, physicsClientId=self.id)

    def disable_joint_force_torque_sensor(self, obj: Object, joint_id: int) -> None:
        p.enableJointForceTorqueSensor(obj.id, joint_id, enableSensor=0, physicsClientId=self.id)

    def get_joint_reaction_force_torque(self, obj: Object, joint_id: int) -> List[float]:
        return p.getJointState(obj.id, joint_id, physicsClientId=self.id)[2]

    def get_applied_joint_motor_torque(self, obj: Object, joint_id: int) -> float:
        return p.getJointState(obj.id, joint_id, physicsClientId=self.id)[3]


class Gui(threading.Thread):
    """
    For internal use only. Creates a new thread for the physics simulation that is active until closed by
    :func:`~World.exit`
    Also contains the code for controlling the camera.
    """

    def __init__(self, world: World, mode: WorldMode):
        threading.Thread.__init__(self)
        self.world = world
        self.mode: WorldMode = mode
        self.camera_button_id = -1

        # Checks if there is a display connected to the system. If not, the simulation will be run in direct mode.
        if "DISPLAY" not in os.environ:
            loginfo("No display detected. Running the simulation in direct mode.")
            self.mode = WorldMode.DIRECT

    def run(self):
        """
        Initializes the new simulation and checks in an endless loop
        if it is still active. If it is the thread will be suspended for 1/80 seconds, if it is not the method and
        thus the thread terminates. The loop also checks for mouse and keyboard inputs to control the camera.
        """
        if self.mode == WorldMode.DIRECT:
            self.world.id = p.connect(p.DIRECT)
        else:
            self.world.id = p.connect(p.GUI)
            self.camera_button_id = p.addUserDebugParameter("Save as Default Camera", 1, 0, 1, physicsClientId=self.world.id)

            # DISCLAIMER
            # This camera control only works if the WorldMode.GUI BulletWorld is the first one to be created. This is
            # due to a bug in the function pybullet.getDebugVisualizerCamera() which only returns the information of
            # the first created simulation.

            # Disable the side windows of the GUI
            p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0, physicsClientId=self.world.id)
            # Change the init camera pose
            default_camera_config = WorldConfig.default_camera_config
            p.resetDebugVisualizerCamera(cameraDistance=default_camera_config["dist"],
                                         cameraYaw=default_camera_config["yaw"],
                                         cameraPitch=default_camera_config["pitch"],
                                         cameraTargetPosition=default_camera_config["target_position"],
                                         physicsClientId=self.world.id)

            # Get the initial camera target location
            camera_target_position = p.getDebugVisualizerCamera(physicsClientId=self.world.id)[11]

            sphere_visual_id = p.createVisualShape(p.GEOM_SPHERE, radius=0.05, rgbaColor=[1, 0, 0, 1],
                                                   physicsClientId=self.world.id)

            # Create a sphere with a radius of 0.05 and a mass of 0
            sphere_uid = p.createMultiBody(baseMass=0.0,
                                           baseInertialFramePosition=[0, 0, 0],
                                           baseVisualShapeIndex=sphere_visual_id,
                                           basePosition=camera_target_position,
                                           physicsClientId=self.world.id)

            # Define the maxSpeed, used in calculations
            max_speed = 16

            # Set initial Camera Rotation
            camera_yaw = default_camera_config["yaw"]
            camera_pitch = default_camera_config["pitch"]

            # Keep track of the mouse state
            mouse_state = [0, 0, 0]
            old_mouse_x, old_mouse_y = 0, 0

            # Determines if the sphere at cameraTargetPosition is visible
            visible = 1

            # Initial value for the camera button
            last_button_value = 1

            # Loop to update the camera position based on keyboard events
            while p.isConnected(self.world.id):
                # Check if Camera button was pressed
                camera_button_value = p.readUserDebugParameter(self.camera_button_id)
                if camera_button_value != last_button_value:
                    last_button_value = camera_button_value

                    current_camera_config = p.getDebugVisualizerCamera()[8:]
                    v = dict(zip(["yaw", "pitch", "dist", "target_position"], current_camera_config))
                    v["target_position"] = list(v["target_position"])
                    yaml_path = os.path.join(os.path.dirname(__file__), "..", "config", 'camera.yaml')
                    with open(yaml_path, "w") as f:
                        yaml.dump(v, f)


                # Monitor user input
                keys = p.getKeyboardEvents(self.world.id)
                mouse = p.getMouseEvents(self.world.id)

                # Get infos about the camera
                width, height, dist = (p.getDebugVisualizerCamera()[0],
                                       p.getDebugVisualizerCamera()[1],
                                       p.getDebugVisualizerCamera()[10])
                # print("width: ", width, "height: ", height, "dist: ", dist)
                camera_target_position = p.getDebugVisualizerCamera(self.world.id)[11]

                # Get vectors used for movement on x,y,z Vector
                x_vec = [p.getDebugVisualizerCamera(self.world.id)[2][i] for i in [0, 4, 8]]
                y_vec = [p.getDebugVisualizerCamera(self.world.id)[2][i] for i in [2, 6, 10]]
                z_vec = (0, 0, 1)  # [p.getDebugVisualizerCamera()[2][i] for i in [1, 5, 9]]

                # Check the mouse state
                if mouse:
                    for m in mouse:

                        mouse_x = m[2]
                        mouse_y = m[1]

                        # update mouseState
                        # Left Mouse button
                        if m[0] == 2 and m[3] == 0:
                            mouse_state[0] = m[4]
                        # Middle mouse button (scroll wheel)
                        if m[0] == 2 and m[3] == 1:
                            mouse_state[1] = m[4]
                        # right mouse button
                        if m[0] == 2 and m[3] == 2:
                            mouse_state[2] = m[4]

                        # change visibility by clicking the mousewheel
                        # if m[4] == 6 and m[3] == 1 and visible == 1:
                        #     visible = 0
                        # elif m[4] == 6 and visible == 0:
                        #     visible = 1

                        # camera movement when the left mouse button is pressed
                        if mouse_state[0] == 3:
                            speed_x = abs(old_mouse_x - mouse_x) if (abs(old_mouse_x - mouse_x)) < max_speed \
                                else max_speed
                            speed_y = abs(old_mouse_y - mouse_y) if (abs(old_mouse_y - mouse_y)) < max_speed \
                                else max_speed

                            # max angle of 89.5 and -89.5 to make sure the camera does not flip (is annoying)
                            if mouse_x < old_mouse_x:
                                if (camera_pitch + speed_x) < 89.5:
                                    camera_pitch += (speed_x / 4) + 1
                            elif mouse_x > old_mouse_x:
                                if (camera_pitch - speed_x) > -89.5:
                                    camera_pitch -= (speed_x / 4) + 1

                            if mouse_y < old_mouse_y:
                                camera_yaw += (speed_y / 4) + 1
                            elif mouse_y > old_mouse_y:
                                camera_yaw -= (speed_y / 4) + 1

                        # Camera movement when the middle mouse button is pressed
                        if mouse_state[1] == 3:
                            speed_x = abs(old_mouse_x - mouse_x)
                            factor = 0.05

                            if mouse_x < old_mouse_x:
                                dist = dist - speed_x * factor
                            elif mouse_x > old_mouse_x:
                                dist = dist + speed_x * factor
                            dist = max(dist, 0.1)

                        # camera movement when the right mouse button is pressed
                        if mouse_state[2] == 3:
                            speed_x = abs(old_mouse_x - mouse_x) if (abs(old_mouse_x - mouse_x)) < 5 else 5
                            speed_y = abs(old_mouse_y - mouse_y) if (abs(old_mouse_y - mouse_y)) < 5 else 5
                            factor = 0.05

                            if mouse_x < old_mouse_x:
                                camera_target_position = np.subtract(camera_target_position,
                                                                     np.multiply(np.multiply(z_vec, factor), speed_x))
                            elif mouse_x > old_mouse_x:
                                camera_target_position = np.add(camera_target_position,
                                                                np.multiply(np.multiply(z_vec, factor), speed_x))

                            if mouse_y < old_mouse_y:
                                camera_target_position = np.add(camera_target_position,
                                                                np.multiply(np.multiply(x_vec, factor), speed_y))
                            elif mouse_y > old_mouse_y:
                                camera_target_position = np.subtract(camera_target_position,
                                                                     np.multiply(np.multiply(x_vec, factor), speed_y))
                        # update oldMouse values
                        old_mouse_y, old_mouse_x = mouse_y, mouse_x

                # check the keyboard state
                if keys:
                    # if shift is pressed, double the speed
                    if p.B3G_SHIFT in keys:
                        speed_mult = 5
                    else:
                        speed_mult = 2.5

                    # if control is pressed, the movements caused by the arrowkeys, the '+' as well as the '-' key
                    # change
                    if p.B3G_CONTROL in keys:

                        # the up and down arrowkeys cause the targetPos to move along the z axis of the map
                        if p.B3G_DOWN_ARROW in keys:
                            camera_target_position = np.subtract(camera_target_position,
                                                                 np.multiply(np.multiply(z_vec, 0.03), speed_mult))
                        elif p.B3G_UP_ARROW in keys:
                            camera_target_position = np.add(camera_target_position,
                                                            np.multiply(np.multiply(z_vec, 0.03), speed_mult))

                        # left and right arrowkeys cause the targetPos to move horizontally relative to the camera
                        if p.B3G_LEFT_ARROW in keys:
                            camera_target_position = np.subtract(camera_target_position,
                                                                 np.multiply(np.multiply(x_vec, 0.03), speed_mult))
                        elif p.B3G_RIGHT_ARROW in keys:
                            camera_target_position = np.add(camera_target_position,
                                                            np.multiply(np.multiply(x_vec, 0.03), speed_mult))

                        # the '+' and '-' keys cause the targetpos to move forwards and backwards relative to the camera
                        # while the camera stays at a constant distance. SHIFT + '=' is for US layout
                        if ord("+") in keys or p.B3G_SHIFT in keys and ord("=") in keys:
                            camera_target_position = np.subtract(camera_target_position,
                                                                 np.multiply(np.multiply(y_vec, 0.03), speed_mult))
                        elif ord("-") in keys:
                            camera_target_position = np.add(camera_target_position,
                                                            np.multiply(np.multiply(y_vec, 0.03), speed_mult))

                    # standard bindings for thearrowkeys, the '+' as well as the '-' key
                    else:

                        # left and right arrowkeys cause the camera to rotate around the yaw axis
                        if p.B3G_RIGHT_ARROW in keys:
                            camera_yaw += (360 / width) * speed_mult
                        elif p.B3G_LEFT_ARROW in keys:
                            camera_yaw -= (360 / width) * speed_mult

                        # the up and down arrowkeys cause the camera to rotate around the pitch axis
                        if p.B3G_DOWN_ARROW in keys:
                            if (camera_pitch + (360 / height) * speed_mult) < 89.5:
                                camera_pitch += (360 / height) * speed_mult
                        elif p.B3G_UP_ARROW in keys:
                            if (camera_pitch - (360 / height) * speed_mult) > -89.5:
                                camera_pitch -= (360 / height) * speed_mult

                        # the '+' and '-' keys cause the camera to zoom towards and away from the targetPos without
                        # moving it. SHIFT + '=' is for US layout since the events can't handle shift plus something
                        if ord("+") in keys or p.B3G_SHIFT in keys and ord("=") in keys:
                            if (dist - (dist * 0.02) * speed_mult) > 0.1:
                                dist -= dist * 0.02 * speed_mult
                        elif ord("-") in keys:
                            dist += dist * 0.02 * speed_mult
                # print("dist: ", dist)
                # print("camera_yaw: ", camera_yaw)
                # print("camera_pitch: ", camera_pitch)
                # print("camera_target_position: ", camera_target_position)

                p.resetDebugVisualizerCamera(cameraDistance=dist, cameraYaw=camera_yaw, cameraPitch=camera_pitch,
                                             cameraTargetPosition=camera_target_position, physicsClientId=self.world.id)
                if visible == 0:
                    camera_target_position = (0.0, -50, 50)
                p.resetBasePositionAndOrientation(sphere_uid, camera_target_position, [0, 0, 0, 1],
                                                  physicsClientId=self.world.id)
                time.sleep(1. / 80.)



import atexit
from datetime import timedelta
import time

import tf
from geometry_msgs.msg import TransformStamped
from sensor_msgs.msg import JointState
from typing_extensions import Optional

from ..datastructures.world import World
from ..robot_description import RobotDescription
from ..datastructures.pose import PoseStamped
from ..ros import  Time, Duration
from ..ros import  wait_for_message, create_timer


class WorldStateUpdater:
    """
    Updates the robot in the World with information of the real robot published to ROS topics.
    Infos used to update the robot are:

        * The current pose of the robot
        * The current joint state of the robot
    """

    def __init__(self, tf_topic: str, joint_state_topic: str, update_rate: timedelta = timedelta(milliseconds=100),
                 world: Optional[World] = None) -> None:
        """
        The robot state updater uses a TF topic and a joint state topic to get the current state of the robot.

        :param tf_topic: Name of the TF topic, needs to publish geometry_msgs/TransformStamped
        :param joint_state_topic: Name of the joint state topic, needs to publish sensor_msgs/JointState
        """
        self.tf_listener = tf.TransformListener()
        time.sleep(1)
        self.tf_topic = tf_topic
        self.joint_state_topic = joint_state_topic
        self.world: Optional[World] = world
        self.tf_timer = create_timer(Duration().from_sec(update_rate.total_seconds()), self._subscribe_tf)
        self.joint_state_timer = create_timer(Duration().from_sec(update_rate.total_seconds()),
                                              self._subscribe_joint_state)

        atexit.register(self._stop_subscription)

    def _subscribe_tf(self, msg: TransformStamped) -> None:
        """
        Callback for the TF timer, will do a lookup of the transform between map frame and the objects frames.

        :param msg: TransformStamped message published to the topic
        """
        if self.world is None:
            if not World.current_world.is_prospection_world:
                self.world = World.current_world
            else:
                return
        for obj in self.world.objects:
            if obj.name == self.world.robot.name:
                tf_frame = RobotDescription.current_robot_description.base_link
            elif obj.is_an_environment:
                continue
            elif obj.is_an_object:
                continue
            else:
                tf_frame = obj.tf_frame
            trans, rot = self.tf_listener.lookupTransform("/map", tf_frame, Time(0))
            obj.set_pose(PoseStamped.from_list(trans, rot))

    def _subscribe_joint_state(self, msg: JointState) -> None:
        """
        Sets the current joint configuration of the robot in the world to the configuration published on the
        topic. Since this uses rospy.wait_for_message which can have errors when used with threads there might be an
        attribute error in the rospy implementation.

        :param msg: JointState message published to the topic.
        """
        try:
            msg = wait_for_message(self.joint_state_topic, JointState)
            joint_positions = dict(zip(msg.name, msg.position))
            World.robot.set_multiple_joint_positions(joint_positions)
        except AttributeError:
            pass

    def _stop_subscription(self) -> None:
        """
        Stops the Timer for TF and joint states and therefore the updating of the robot in the world.
        """
        self.tf_timer.shutdown()
        self.joint_state_timer.shutdown()




import time
import rclpy.client

from . import node
from .logging import loginfo_once

services = {}

class ServiceProxy:
    def __init__(self, topic_name, service_message):
        self.service = node.create_client(service_message, topic_name)
        self.message_type = service_message
    def __call__(self, *args, **kwargs):
        future = self.service.call_async(*args, **kwargs)
        while not future.done():
            time.sleep(0.1)
        return future.result()

    def wait_for_service(self, *args, **kwargs):
        self.service.wait_for_service()


def get_service_proxy(topic_name: str, service_message) -> rclpy.client.Client:
    service = ServiceProxy(topic_name, service_message)
    services[topic_name] = service
    return service

def wait_for_service(topic_name: str, service_message):
    if not topic_name in services:
        service = get_service_proxy(topic_name, service_message)
    loginfo_once(f"Waiting for service: {topic_name}")
    services[topic_name].wait_for_service()